# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by debusine tasks."""

import datetime as dt
import re
from abc import ABC
from collections.abc import Collection, Iterator, Sequence
from enum import StrEnum
from functools import partial
from itertools import chain, groupby
from pathlib import Path
from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
    Self,
    TypeAlias,
    Union,
    override,
)

import debian.deb822 as deb822
import pydantic

from debusine.artifacts.models import (
    ARTIFACT_DATA_MODELS_BY_CATEGORY,
    ArtifactCategory,
    ArtifactData,
    BareDataCategory,
    CollectionCategory,
    DebianPackageBuildLog,
    RuntimeStatistics,
    WorkRequestResults,
)
from debusine.client.models import LookupChildType
from debusine.utils import DjangoChoicesEnum


class WorkerType(DjangoChoicesEnum):
    """The type of a Worker."""

    SCHEDULER = "scheduler"
    EXTERNAL = "external"
    CELERY = "celery"
    SIGNING = "signing"


class BackendType(StrEnum):
    """Possible values for backend."""

    AUTO = "auto"
    UNSHARE = "unshare"
    INCUS_LXC = "incus-lxc"
    INCUS_VM = "incus-vm"
    QEMU = "qemu"


class AutopkgtestNeedsInternet(StrEnum):
    """Possible values for needs_internet."""

    RUN = "run"
    TRY = "try"
    SKIP = "skip"


class LintianFailOnSeverity(StrEnum):
    """Possible values for fail_on_severity."""

    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    PEDANTIC = "pedantic"
    EXPERIMENTAL = "experimental"
    OVERRIDDEN = "overridden"
    NONE = "none"


class BaseTaskDataModel(pydantic.BaseModel):
    """Stricter pydantic defaults for task data models."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    # Replace with serialize_by_alias=True in model_config once we can
    # assume pydantic >= 2.11.
    @override
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Use aliases by default when serializing."""
        kwargs.setdefault("by_alias", True)
        return super().model_dump(**kwargs)


# https://github.com/pydantic/pydantic-core/issues/296
def require_unique(v: list[Any]) -> list[Any]:
    """Field validator to require that list items are unique."""
    if sorted(v) != sorted(set(v)):
        raise ValueError(f"{v} contains duplicate items")
    return v


# Lookups resulting in exactly one collection item.
LookupSingle: TypeAlias = int | str


def _build_key_value_string(data: dict[str, Any]) -> str:
    # TODO: Implement URL-quoting to support special characters
    return ":".join(f"{key}={value}" for key, value in data.items())


def build_lookup_string(
    *segments: str, options: dict[str, Any] | None = None
) -> str:
    """Build a string lookup from segments and options."""
    options_prefix = ""
    if options:
        inside = _build_key_value_string(options)
        options_prefix = f"//{inside}/"
    return f"{options_prefix}{"/".join(segments)}"


def _parse_key_value_string(input_string: str) -> dict[str, str]:
    data = {}
    for item in input_string.split(":"):
        # TODO: Implement URL-unquoting to support special characters
        name, value = item.split("=", 1)
        data[name] = value
    return data


def parse_lookup_string(lookup: str) -> tuple[list[str], dict[str, str] | None]:
    """Parse a string lookup into options and segments."""
    options = None

    # Extract and parse options
    if m := re.match(r"//([^/]*)/", lookup):
        lookup_options = m.group(1)
        lookup = lookup[len(lookup_options) + 3 :]
        options = _parse_key_value_string(lookup_options)

    return (lookup.split("/"), options)


def build_key_value_lookup_segment(
    lookup_type: str, filters: dict[str, str]
) -> str:
    """Build a lookup segment consisting of a series of `key=value` filters."""
    filters_string = _build_key_value_string(filters)
    return f"{lookup_type}:{filters_string}"


def parse_key_value_lookup_segment(segment: str) -> tuple[str, dict[str, str]]:
    """Parse a lookup segment consisting of a series of `key=value` filters."""
    lookup_type, remaining = segment.split(":", 1)
    return lookup_type, _parse_key_value_string(remaining)


class CollectionItemMatcherKind(StrEnum):
    """Possible values for CollectionItemMatcher.kind."""

    CONTAINS = "contains"
    ENDSWITH = "endswith"
    EXACT = "exact"
    IN = "in"
    STARTSWITH = "startswith"


class CollectionItemMatcher(BaseTaskDataModel):
    """A matcher for collection item name or per-item data fields."""

    model_config = BaseTaskDataModel.model_config | pydantic.ConfigDict(
        frozen=True
    )

    kind: CollectionItemMatcherKind
    value: Any


class BaseExtraRepository(BaseTaskDataModel, ABC):
    """Generic extra repositories for Debian tasks."""

    components: (
        list[Annotated[str, pydantic.Field(pattern=r"^(\w|-)+$")]] | None
    ) = None

    architectures: (
        list[Annotated[str, pydantic.Field(pattern=r"^(\w|-)+$")]] | None
    ) = None

    @pydantic.field_validator("components", mode="after")
    @classmethod
    def convert_empty_components_to_none(
        cls, value: list[str] | None
    ) -> list[str] | None:
        """Rewrite an empty components list to None."""
        return value or None

    @pydantic.field_validator("architectures", mode="after")
    @classmethod
    def convert_empty_architectures_to_none(
        cls, value: list[str] | None
    ) -> list[str] | None:
        """Rewrite an empty architectures list to None."""
        return value or None


class ExtraDebusineRepository(BaseExtraRepository):
    """An extra Debusine-hosted repository for Debian tasks."""

    suite: LookupSingle


class ExtraExternalRepository(BaseExtraRepository):
    """An extra external repository for Debian tasks."""

    url: Annotated[pydantic.AnyUrl, pydantic.UrlConstraints(host_required=True)]
    suite: Annotated[str, pydantic.Field(pattern=r"^(\w|[./ -])+$")]
    signing_key: str | None = None

    @pydantic.model_validator(mode="after")
    def flat_repository(self) -> Self:
        """Check if suite is a flat repository."""
        if self.suite.endswith("/"):
            if self.components is not None:
                raise ValueError(
                    "Components cannot be specified for a flat "
                    "repository (where the suite ends with /)"
                )
        elif not self.components:
            raise ValueError("Components must be specified")
        return self

    def as_deb822_source(self, signed_by_filename: str | None = None) -> str:
        """Render a Deb822 sources.list.d entry."""
        source = deb822.Deb822()
        source["Types"] = "deb"
        source["URIs"] = str(self.url)
        source["Suites"] = self.suite
        if self.components:
            source["Components"] = " ".join(self.components)
        if self.signing_key:
            if signed_by_filename:
                source["Signed-By"] = signed_by_filename
            else:
                source["Signed-By"] = "\n" + "\n".join(
                    " " + (line or ".")
                    for line in self.signing_key.splitlines()
                )
        return source.dump()


ExtraRepository: TypeAlias = ExtraDebusineRepository | ExtraExternalRepository


class LookupDict(BaseTaskDataModel):
    """Dictionary lookups for collection items."""

    model_config = BaseTaskDataModel.model_config | pydantic.ConfigDict(
        frozen=True
    )

    collection: LookupSingle
    child_type: LookupChildType = LookupChildType.ARTIFACT
    category: str | None = None
    name_matcher: CollectionItemMatcher | None = None
    # Logically a mapping, but we turn it into a tuple so that lookups are
    # hashable.
    data_matchers: tuple[tuple[str, CollectionItemMatcher], ...] = ()
    # Logically a mapping, but we turn it into a tuple so that lookups are
    # hashable.  In future some filters may want other value types, but they
    # need to be hashable and our current use cases involve subordinate
    # lookups, so just declare them to be lookups for now.
    lookup_filters: tuple[
        tuple[str, Union[LookupSingle, "LookupMultiple"]], ...
    ] = ()
    error_on_empty: bool = False

    @pydantic.model_validator(mode="before")
    @classmethod
    def normalize_matchers(cls, data: Any) -> dict[str, Any]:  # noqa: C901
        """
        Transform the lookup syntax into a form more convenient for pydantic.

        `name`, `name__*`, `data__KEY`, and `data__KEY__*` keys in `values`
        are transformed into :py:class:`CollectionItemMatcher` instances
        stored in `name_matcher` and `data_matchers`.  `lookup__KEY` keys
        are transformed into entries in `lookup_filters`.  Conflicting
        lookup suffixes are rejected.
        """

        def split_words(s: str) -> list[str]:
            return s.split("__")

        def matcher_prefix(s: str) -> tuple[str, ...]:
            match split_words(s):
                case ["name", *_]:
                    return ("name",)
                case ["data", data_key, *_]:
                    return ("data", data_key)
                case ["lookup", lookup_key]:
                    return ("lookup", lookup_key)
                case _:
                    return ()

        if not isinstance(data, dict):
            raise ValueError(f"Expected dict, got {type(data)!r}")
        data = data.copy()
        assert isinstance(data, dict)

        data_matchers: dict[str, CollectionItemMatcher] = dict(
            data.get("data_matchers", ())
        )
        lookup_filters: dict[str, LookupSingle | LookupMultiple] = dict(
            data.get("lookup_filters", ())
        )

        for key, group in groupby(
            sorted(data, key=matcher_prefix), matcher_prefix
        ):
            if not key:
                # Only keys with `name`, `data`, or `lookup` as their first
                # segment need transformation.
                continue
            matcher_names = list(group)
            if len(matcher_names) > 1:
                raise ValueError(
                    f"Conflicting matchers: {sorted(matcher_names)}"
                )
            assert len(matcher_names) == 1
            matcher_name = matcher_names[0]

            data_key: str | None = None
            match words := split_words(matcher_name):
                case ["name"]:
                    kind = "exact"
                case ["name", kind] if kind != "exact":
                    pass
                case ["data", data_key]:
                    kind = "exact"
                case ["data", data_key, kind] if kind != "exact":
                    pass
                case ["lookup", lookup_key]:
                    match data[matcher_name]:
                        case int() | str():
                            lookup_filters[lookup_key] = data[matcher_name]
                        case _:
                            lookup_filters[lookup_key] = (
                                LookupMultiple.model_validate(
                                    data[matcher_name]
                                )
                            )
                    for k in matcher_names:
                        del data[k]
                    continue
                case _:
                    # We filter out keys with first segments other than
                    # `name`, `data`, or `lookup` earlier, so anything that
                    # reaches here must be a malformed lookup.
                    raise ValueError(f"Unrecognized matcher: {matcher_name}")

            matcher = CollectionItemMatcher(
                kind=CollectionItemMatcherKind(kind), value=data[matcher_name]
            )
            if words[0] == "name":
                data["name_matcher"] = matcher
            else:
                assert words[0] == "data"
                assert data_key is not None
                data_matchers[data_key] = matcher
            for k in matcher_names:
                del data[k]

        if data_matchers:
            data["data_matchers"] = tuple(sorted(data_matchers.items()))
        if lookup_filters:
            data["lookup_filters"] = tuple(sorted(lookup_filters.items()))

        return data

    def export(self) -> dict[str, Any]:
        """
        Export the usual input representation of this lookup.

        This reverses the transformations applied by
        :py:meth:`normalize_matchers`.
        """
        value: dict[str, Any] = {"collection": self.collection}
        if self.child_type != LookupChildType.ARTIFACT:
            value["child_type"] = str(self.child_type)
        if self.category is not None:
            value["category"] = self.category
        if self.name_matcher is not None:
            if self.name_matcher.kind == CollectionItemMatcherKind.EXACT:
                value["name"] = self.name_matcher.value
            else:
                value[f"name__{self.name_matcher.kind}"] = (
                    self.name_matcher.value
                )
        for key, matcher in self.data_matchers:
            if matcher.kind == CollectionItemMatcherKind.EXACT:
                value[f"data__{key}"] = matcher.value
            else:
                value[f"data__{key}__{matcher.kind}"] = matcher.value
        for key, filter_value in self.lookup_filters:
            if isinstance(filter_value, LookupMultiple):
                value[f"lookup__{key}"] = filter_value.export()
            else:
                value[f"lookup__{key}"] = filter_value
        if self.error_on_empty:
            value["error_on_empty"] = True
        return value


class LookupMultiple(pydantic.RootModel[tuple[LookupSingle | LookupDict, ...]]):
    """Lookups resulting in multiple collection items."""

    model_config = pydantic.ConfigDict(validate_assignment=True, frozen=True)

    @pydantic.model_validator(mode="before")
    @classmethod
    def normalize(cls, data: Any) -> tuple[Any, ...]:
        """Normalize into a tuple of multiple matchers."""
        if isinstance(data, dict):
            return (data,)
        elif isinstance(data, (tuple, list)):
            return tuple(data)
        else:
            raise ValueError(
                "Lookup of multiple collection items must be a dictionary or "
                "a list"
            )

    @override
    def __iter__(  # type: ignore[override]
        self,
    ) -> Iterator[LookupSingle | LookupDict]:
        """Iterate over individual lookups."""
        return iter(self.root)

    def __bool__(self) -> bool:
        """Return True if and only if this lookup is non-empty."""
        return bool(self.root)

    def export(self) -> dict[str, Any] | list[int | str | dict[str, Any]]:
        """
        Export the usual input representation of this lookup.

        This reverses the transformations applied by :py:meth:`normalize`.
        """
        if len(self.root) == 1 and isinstance(self.root[0], LookupDict):
            return self.root[0].export()
        else:
            return [
                lookup.export() if isinstance(lookup, LookupDict) else lookup
                for lookup in self
            ]


# Resolve circular reference to LookupMultiple.
LookupDict.model_rebuild()


class NotificationDataEmail(BaseTaskDataModel):
    """Channel data for email notifications."""

    from_: pydantic.EmailStr | None = pydantic.Field(default=None, alias="from")
    to: list[pydantic.EmailStr] | None = None
    cc: list[pydantic.EmailStr] = pydantic.Field(default_factory=list)
    subject: str | None = None


class ActionTypes(StrEnum):
    """Possible values for EventReaction actions."""

    SKIP_IF_LOOKUP_RESULT_CHANGED = "skip-if-lookup-result-changed"
    SEND_NOTIFICATION = "send-notification"
    UPDATE_COLLECTION_WITH_ARTIFACTS = "update-collection-with-artifacts"
    UPDATE_COLLECTION_WITH_DATA = "update-collection-with-data"
    RETRY_WITH_DELAYS = "retry-with-delays"
    RECORD_IN_TASK_HISTORY = "record-in-task-history"


class ActionSkipIfLookupResultChanged(BaseTaskDataModel):
    """Action for skipping a work request if a lookup result changed."""

    action: Literal[ActionTypes.SKIP_IF_LOOKUP_RESULT_CHANGED] = (
        ActionTypes.SKIP_IF_LOOKUP_RESULT_CHANGED
    )
    # Deliberately not LookupSingle; this must be able to return a
    # collection item, so integer lookups are never valid.
    lookup: str
    collection_item_id: int | None
    promise_name: str | None = None


class ActionSendNotification(BaseTaskDataModel):
    """Action for sending a notification."""

    action: Literal[ActionTypes.SEND_NOTIFICATION] = (
        ActionTypes.SEND_NOTIFICATION
    )
    channel: str
    data: NotificationDataEmail | None = None


class ActionUpdateCollectionWithArtifacts(BaseTaskDataModel):
    """Action for updating a collection with artifacts."""

    action: Literal[ActionTypes.UPDATE_COLLECTION_WITH_ARTIFACTS] = (
        ActionTypes.UPDATE_COLLECTION_WITH_ARTIFACTS
    )
    collection: LookupSingle
    name_template: str | None = None
    variables: dict[str, Any] | None = None
    artifact_filters: dict[str, Any]
    created_at: dt.datetime | None = None


class ActionUpdateCollectionWithData(BaseTaskDataModel):
    """Action for updating a collection with bare data."""

    action: Literal[ActionTypes.UPDATE_COLLECTION_WITH_DATA] = (
        ActionTypes.UPDATE_COLLECTION_WITH_DATA
    )
    collection: LookupSingle
    category: BareDataCategory
    name_template: str | None = None
    data: dict[str, Any] | None = None
    created_at: dt.datetime | None = None


class ActionRetryWithDelays(BaseTaskDataModel):
    """Action for retrying a work request with delays."""

    _delay_re: ClassVar[re.Pattern[str]] = re.compile(r"^([0-9]+)([mhdw])$")

    action: Literal[ActionTypes.RETRY_WITH_DELAYS] = (
        ActionTypes.RETRY_WITH_DELAYS
    )
    delays: Annotated[
        list[Annotated[str, pydantic.Field(pattern=_delay_re)]],
        (pydantic.Field(min_length=1)),
    ]


class ActionRecordInTaskHistory(BaseTaskDataModel):
    """Action for recording the task run in a task-history collection."""

    action: Literal[ActionTypes.RECORD_IN_TASK_HISTORY] = (
        ActionTypes.RECORD_IN_TASK_HISTORY
    )
    subject: str | None = None
    context: str | None = None


EventReaction = Annotated[
    ActionSkipIfLookupResultChanged
    | ActionSendNotification
    | ActionUpdateCollectionWithArtifacts
    | ActionUpdateCollectionWithData
    | ActionRetryWithDelays
    | ActionRecordInTaskHistory,
    pydantic.Field(discriminator="action"),
]


class EventReactions(BaseTaskDataModel):
    """Structure for event reactions."""

    on_creation: list[EventReaction] = []
    on_unblock: list[EventReaction] = []
    on_assignment: list[EventReaction] = []
    on_success: list[EventReaction] = []
    on_failure: list[EventReaction] = []

    @override
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Wrap pydantic to ensure discriminators are serializers."""
        # TODO: from python 3.5, docstrings are inherited from the parent
        # class, which would be the right thing here. However, flake8 doesn't
        # like it.

        # Mark discriminators as always set.
        # See: https://github.com/pydantic/pydantic/issues/6465
        for name in (
            "on_creation",
            "on_unblock",
            "on_assignment",
            "on_success",
            "on_failure",
        ):
            for item in getattr(self, name):
                item.model_fields_set.add("action")

        return super().model_dump(**kwargs)

    def add_reaction(
        self,
        name: Literal[
            "on_creation",
            "on_unblock",
            "on_assignment",
            "on_success",
            "on_failure",
        ],
        action: EventReaction,
    ) -> None:
        """
        Add an EventReaction to the named event.

        This is a noop if the named event already contains this EventReaction.
        """
        current = getattr(self, name)
        if action not in current:
            current.append(action)
            # Let pydantic know we modified the field.
            # See: https://github.com/pydantic/pydantic/issues/9866
            self.model_fields_set.add(name)


class BaseTaskData(BaseTaskDataModel):
    """
    Base class for task data.

    Task data is encoded as JSON in the database and in the API, and it is
    modeled as a pydantic data structure in memory for both ease of access and
    validation.
    """

    DEFAULT_TASK_CONFIGURATION_LOOKUP: ClassVar[str] = (
        "default@debusine:task-configuration"
    )

    #: debusine:task-configuration collection to use to configure the task.
    #: This is resolved to an ID when configuration is applied
    task_configuration: LookupSingle | None = DEFAULT_TASK_CONFIGURATION_LOOKUP


class BaseTaskDataWithExecutor(BaseTaskData):
    """Base task data with fields used to configure executors."""

    backend: BackendType = BackendType.AUTO
    environment: LookupSingle | None = None

    # Not all executor-based tasks strictly require this, but it can
    # nevertheless be useful to constrain a task to run on an architecture
    # for which we know that a suitable environment exists.
    build_architecture: str | None = None


class BaseTaskDataWithExtraRepositories(BaseTaskData):
    """Base task data with fields used to configure extra_repositories."""

    extra_repositories: list[ExtraRepository] | None = None
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    build_architecture: str


class BaseDynamicTaskData(BaseTaskDataModel):
    """
    Base class for dynamic task data.

    This is computed by the scheduler when dispatching a task to a worker.
    It may involve resolving artifact lookups.
    """

    #: Brief human-readable summary of the most important parameters to this
    #: work request.
    parameter_summary: str | None = None

    #: An abstract string value representing the subject of the task, for
    #: the purpose of recording statistics.  It is meant to group possible
    #: inputs into groups that we expect to behave similarly.
    subject: str | None = None

    #: An abstract string value representing the runtime context in which
    #: the task is executed, for the purpose of recording statistics.  It is
    #: meant to represent some of the task parameters that can significantly
    #: alter the runtime behaviour of the task.
    runtime_context: str | None = None

    #: Name of the configuration context
    # This is a stable string representation of the main input artifacts and
    # parameters, that does not change across similar future invocations of the
    # same task, and is used to look up configuration parameters.
    configuration_context: str | None = None

    #: Storage for resolved task input fields
    task_input_fields: dict[str, Any] = pydantic.Field(default_factory=dict)

    def get_source_package_name(self) -> str | None:
        """
        Extract the source package name.

        :returns: the source package name, if applicable, and None otherwise.
        """
        return None

    def get_input_artifacts_ids(self) -> list[int]:
        """
        Return the list of input artifact IDs used by this task.

        This is used by views to show what artifacts were used by a task.
        """
        # BaseExternalTask.get_input_artifacts_ids() augments this with
        # artifact IDs extracted from resolved task input fields, which are
        # not serialized in old dynamic_data records.
        return []


class BaseDynamicTaskDataWithExecutor(BaseDynamicTaskData):
    """Dynamic task data for executors."""

    environment_id: int | None = None


class BaseDynamicTaskDataWithExtraRepositories(BaseDynamicTaskData):
    """Dynamic task data for extra repositories."""

    extra_repositories: list[ExtraExternalRepository] | None = None


class OutputDataError(BaseTaskDataModel):
    """An error encountered when running a task."""

    message: str
    code: str


class RegressionAnalysisStatus(StrEnum):
    """Possible values for ``RegressionAnalysis.status``."""

    NO_RESULT = "no-result"
    ERROR = "error"
    IMPROVEMENT = "improvement"
    STABLE = "stable"
    REGRESSION = "regression"


class RegressionAnalysis(BaseTaskDataModel):
    """The result of a regression analysis."""

    original_source_version: str | None = None
    original_artifact_id: int | None = None
    new_source_version: str | None = None
    new_artifact_id: int | None = None
    status: RegressionAnalysisStatus
    details: dict[str, Any] | list[Any] | None = None

    # Obsolete, but may be present in old rows.
    original_url: str | None = None
    new_url: str | None = None


class Confirmation(BaseTaskDataModel):
    """A record that a work request was confirmed or denied by a user."""

    confirmed: bool
    comment: str | None = None


class OutputData(BaseTaskDataModel):
    """Data produced when a task is completed."""

    runtime_statistics: RuntimeStatistics | None = None
    errors: list[OutputDataError] | None = None
    skip_reason: str | None = None
    regression_analysis: dict[str, RegressionAnalysis] | None = None
    confirmation: Confirmation | None = None

    def merge(self, other: "OutputData") -> "OutputData":
        """
        Return the combination of two ``OutputData`` objects.

        Merge all the fields that were explicitly set on each model.  If a
        field is set on both models, the values in ``other`` take
        precedence.
        """
        merged = self.model_copy()
        for name in other.model_fields_set:
            setattr(merged, name, getattr(other, name))
        return merged


class NoopData(BaseTaskDataWithExecutor):
    """In memory task data for the Noop task."""

    result: WorkRequestResults = WorkRequestResults.SUCCESS


class AutopkgtestInput(BaseTaskDataModel):
    """Input for an autopkgtest task."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple
    context_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )


class AutopkgtestFailOn(BaseTaskDataModel):
    """Possible values for fail_on."""

    failed_test: bool = True
    flaky_test: bool = False
    skipped_test: bool = False


class AutopkgtestTimeout(BaseTaskDataModel):
    """Timeout specifications for an autopkgtest task."""

    global_: Annotated[int | None, pydantic.Field(ge=0)] = pydantic.Field(
        default=None, alias="global"
    )
    factor: Annotated[int | None, pydantic.Field(ge=0)] = None
    short: Annotated[int | None, pydantic.Field(ge=0)] = None
    install: Annotated[int | None, pydantic.Field(ge=0)] = None
    test: Annotated[int | None, pydantic.Field(ge=0)] = None
    copy_: Annotated[int | None, pydantic.Field(ge=0)] = pydantic.Field(
        default=None, alias="copy"
    )


class AutopkgtestData(
    BaseTaskDataWithExecutor, BaseTaskDataWithExtraRepositories
):
    """In memory task data for the Autopkgtest task."""

    input: AutopkgtestInput
    include_tests: list[str] = pydantic.Field(default_factory=list)
    exclude_tests: list[str] = pydantic.Field(default_factory=list)
    debug_level: Annotated[int, pydantic.Field(ge=0, le=3)] = 0
    use_packages_from_base_repository: bool = False
    pin_packages: dict[str, list[str]] = pydantic.Field(default_factory=dict)
    extra_environment: dict[str, str] = pydantic.Field(default_factory=dict)
    needs_internet: AutopkgtestNeedsInternet = AutopkgtestNeedsInternet.RUN
    fail_on: AutopkgtestFailOn = pydantic.Field(
        default_factory=AutopkgtestFailOn
    )
    timeout: AutopkgtestTimeout | None = None
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle


class AutopkgtestDynamicData(
    BaseDynamicTaskDataWithExecutor, BaseDynamicTaskDataWithExtraRepositories
):
    """Dynamic data for the Autopkgtest task."""

    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    input_source_artifact_id: int
    input_binary_artifacts_ids: list[int]
    input_context_artifacts_ids: list[int] = pydantic.Field(
        default_factory=list
    )

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [
            self.environment_id,
            self.input_source_artifact_id,
            *self.input_binary_artifacts_ids,
            *self.input_context_artifacts_ids,
        ]


class LintianInput(BaseTaskDataModel):
    """Input for a lintian task."""

    source_artifact: LookupSingle | None = None
    binary_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )

    @pydantic.model_validator(mode="after")
    def check_one_of_source_or_binary(self) -> Self:
        """Ensure a source or binary artifact is present."""
        if self.source_artifact is None and not self.binary_artifacts:
            raise ValueError(
                'One of source_artifact or binary_artifacts must be set'
            )
        return self


class LintianOutput(BaseTaskDataModel):
    """Output configuration for a Lintian task."""

    source_analysis: bool = True
    binary_all_analysis: bool = True
    binary_any_analysis: bool = True


class LintianData(BaseTaskDataWithExecutor):
    """In memory task data for the Lintian task."""

    input: LintianInput
    output: LintianOutput = pydantic.Field(default_factory=LintianOutput)
    target_distribution: str = "debian:unstable"
    # Passed to --tags to lintian
    include_tags: list[str] = pydantic.Field(default_factory=list)
    # Passed to --suppress-tags to lintian
    exclude_tags: list[str] = pydantic.Field(default_factory=list)
    # If the analysis emits tags of this severity or higher, the task will
    # return 'failure' instead of 'success'
    fail_on_severity: LintianFailOnSeverity = LintianFailOnSeverity.ERROR


class LintianDynamicData(BaseDynamicTaskDataWithExecutor):
    """Dynamic data for the Lintian task."""

    input_source_artifact_id: int | None = None
    input_binary_artifacts_ids: list[int] = pydantic.Field(default_factory=list)

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        result = []
        if val := self.input_source_artifact_id:
            result.append(val)
        if val := self.environment_id:
            result.append(val)
        result += self.input_binary_artifacts_ids
        return result


class BlhcInput(BaseTaskDataModel):
    """Input for a blhc task."""

    artifact: LookupSingle


class BlhcFlags(StrEnum):
    """Possible values for extra_flags."""

    ALL = "--all"
    BINDNOW = "--bindnow"
    BUILDD = "--buildd"
    COLOR = "--color"
    DEBIAN = "--debian"
    LINE_NUMBERS = "--line-numbers"
    PIE = "--pie"


class BlhcData(BaseTaskDataWithExecutor):
    """In memory task data for the Blhc task."""

    input: BlhcInput
    # Passed to blhc
    extra_flags: list[BlhcFlags] = pydantic.Field(default_factory=list)

    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    build_architecture: str
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle


class BlhcDynamicData(BaseDynamicTaskDataWithExecutor):
    """Dynamic data for the Blhc task."""

    input_artifact_id: int

    # TODO: Once old tasks created without it have expired, this should
    # become required.
    input_artifact_data: DebianPackageBuildLog | None = None

    # BaseTaskDynamicDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.input_artifact_id, self.environment_id]


class DebDiffInput(BaseTaskDataModel):
    """Input for a debdiff task."""

    source_artifacts: Annotated[
        list[LookupSingle] | None, pydantic.Field(min_length=2, max_length=2)
    ] = None
    binary_artifacts: Annotated[
        list[LookupMultiple] | None, pydantic.Field(min_length=2, max_length=2)
    ] = None

    @pydantic.model_validator(mode="after")
    def check_one_of_source_or_binary(self) -> Self:
        """Ensure only source or binary artifacts are present."""
        if (self.source_artifacts is None) == (self.binary_artifacts is None):
            raise ValueError(
                "Exactly one of source_artifacts"
                " or binary_artifacts must be set"
            )
        return self


class DebDiffFlags(StrEnum):
    """Possible values for extra_flags."""

    DIRS = "--dirs"
    NOCONTROL = "--nocontrol"
    WDIFF = "--wdiff-source-control"
    SHOW_MOVED = "--show-moved"
    DIFFSTAT = "--diffstat"
    PATCHES = "--apply-patches"
    IGNORE_SPACE = "--ignore-space"


class DebDiffData(BaseTaskDataWithExecutor):
    """In memory task data for the DebDiff task."""

    input: DebDiffInput
    # Passed to debdiff
    extra_flags: list[DebDiffFlags] = pydantic.Field(default_factory=list)
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    build_architecture: str
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle


class DebDiffDynamicData(BaseDynamicTaskDataWithExecutor):
    """Dynamic data for the DebDiff task."""

    # BaseTaskDynamicDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    input_source_artifacts_ids: Annotated[
        list[int] | None, pydantic.Field(min_length=2, max_length=2)
    ] = None
    input_binary_artifacts_ids: Annotated[
        list[list[int]] | None, pydantic.Field(min_length=2, max_length=2)
    ] = None

    @pydantic.model_validator(mode="after")
    def validate_source_or_binary_exclusive(self) -> Self:
        """
        Validate mutually exclusive input artifacts.

        Ensure input_source_artifacts_ids and input_binary_artifacts_ids
        are not both set.
        """
        # It currently allows to have "input_source_artifacts_ids" == None and
        # "input_binary_artifacts_ids" == None to allow backwards compatibility
        # with older data
        if (self.input_source_artifacts_ids is not None) and (
            self.input_binary_artifacts_ids is not None
        ):
            raise ValueError(
                "Only one of 'input_source_artifacts_ids' or "
                "'input_binary_artifacts_ids' may be set (not both)."
            )
        return self

    @override
    def get_source_package_name(self) -> str | None:
        if self.subject and self.subject.startswith("source:"):
            return self.subject.removeprefix("source:")
        return None

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        input_ids = self.input_source_artifacts_ids or []
        binary_ids = self.input_binary_artifacts_ids or []
        return list(chain(input_ids, *binary_ids))


class MmDebstrapVariant(StrEnum):
    """Variants supported by `mmdebstrap`."""

    BUILDD = "buildd"
    MINBASE = "minbase"
    DASH = "-"
    APT = "apt"
    CUSTOM = "custom"
    DEBOOTSTRAP = "debootstrap"
    ESSENTIAL = "essential"
    EXTRACT = "extract"
    IMPORTANT = "important"
    REQUIRED = "required"
    STANDARD = "standard"


class DebootstrapVariant(StrEnum):
    """Variants supported by `debootstrap`."""

    BUILDD = "buildd"
    MINBASE = "minbase"


class GpgvCommand(StrEnum):
    """Possible values for gpgv_command."""

    GPGV = "gpgv"
    GPGV_SQ = "gpgv-sq"


class SystemBootstrapOptions(BaseTaskDataModel):
    """Structure of SystemBootstrap options."""

    architecture: str
    # Specializations of this model for individual tasks should restrict
    # this to particular values.
    variant: str | None = None
    extra_packages: list[str] = pydantic.Field(default_factory=list)
    use_signed_by: bool = True
    gpgv_command: GpgvCommand | None = None


class SystemBootstrapRepositoryType(StrEnum):
    """Possible values for repository types."""

    DEB = "deb"
    DEB_SRC = "deb-src"


class SystemBootstrapRepositoryCheckSignatureWith(StrEnum):
    """Possible values for check_signature_with."""

    SYSTEM = "system"
    EXTERNAL = "external"
    NO_CHECK = "no-check"


class SystemBootstrapRepositoryKeyring(BaseTaskDataModel):
    """Description of a repository keyring."""

    url: pydantic.AnyUrl
    sha256sum: str = ""
    install: bool = False

    @pydantic.field_validator("url", mode="after")
    @classmethod
    def validate_url(cls, url: pydantic.AnyUrl) -> pydantic.AnyUrl:
        """
        Reject file:// URLs outside /usr/(local/)share/keyrings/.

        We don't want to allow reading arbitrary paths.
        """
        if (
            url.scheme == "file"
            and url.path is not None
            and not (
                Path(url.path).resolve().is_relative_to("/usr/share/keyrings")
            )
            and not (
                Path(url.path)
                .resolve()
                .is_relative_to("/usr/local/share/keyrings")
            )
        ):
            raise ValueError(
                "file:// URLs for keyrings must be under /usr/share/keyrings/ "
                "or /usr/local/share/keyrings/"
            )
        return url


class SystemBootstrapRepository(BaseTaskDataModel):
    """Description of one repository in SystemBootstrapData."""

    mirror: str
    suite: str
    types: Annotated[
        list[SystemBootstrapRepositoryType],
        pydantic.Field(min_length=1),
        pydantic.AfterValidator(require_unique),
    ] = pydantic.Field(
        default_factory=lambda: [SystemBootstrapRepositoryType.DEB]
    )
    components: (
        Annotated[list[str], pydantic.AfterValidator(require_unique)] | None
    ) = None
    check_signature_with: SystemBootstrapRepositoryCheckSignatureWith = (
        SystemBootstrapRepositoryCheckSignatureWith.SYSTEM
    )
    keyring_package: str | None = None
    keyring: SystemBootstrapRepositoryKeyring | None = None

    @pydantic.model_validator(mode="after")
    def _check_external_keyring(self) -> Self:
        """Require keyring if check_signature_with is external."""
        if (
            self.check_signature_with
            == SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL
        ):
            if self.keyring is None:
                raise ValueError(
                    "repository requires 'keyring': "
                    "'check_signature_with' is set to 'external'"
                )
        return self


class SystemBootstrapData(BaseTaskData):
    """Base for in-memory class data for SystemBootstrap tasks."""

    bootstrap_options: SystemBootstrapOptions
    bootstrap_repositories: Annotated[
        list[SystemBootstrapRepository], pydantic.Field(min_length=1)
    ]
    customization_script: str | None = None


class MmDebstrapBootstrapOptions(SystemBootstrapOptions):
    """Structure of MmDebstrap options."""

    variant: MmDebstrapVariant | None = None
    use_signed_by: bool = True


class MmDebstrapData(SystemBootstrapData):
    """In memory task data for the MmDebstrap task."""

    bootstrap_options: MmDebstrapBootstrapOptions


class DebootstrapBootstrapOptions(SystemBootstrapOptions):
    """Structure of debootstrap options."""

    variant: DebootstrapVariant | None = None


class DiskImageFormat(StrEnum):
    """Possible disk image formats."""

    RAW = "raw"
    QCOW2 = "qcow2"


class Partition(BaseTaskDataModel):
    """Partition definition."""

    size: int
    filesystem: str
    mountpoint: str = "none"


class DiskImage(BaseTaskDataModel):
    """Disk image definition."""

    format: DiskImageFormat
    filename: str = "image"
    kernel_package: str | None = None
    bootloader: str | None = None
    partitions: Annotated[list[Partition], pydantic.Field(min_length=1)]


class SystemImageBuildData(SystemBootstrapData):
    """Base for in-memory class data for SystemImageBuild tasks."""

    bootstrap_options: DebootstrapBootstrapOptions
    disk_image: DiskImage


class PiupartsDataInput(BaseTaskDataModel):
    """Input for a piuparts task."""

    binary_artifacts: LookupMultiple


class PiupartsData(BaseTaskDataWithExecutor, BaseTaskDataWithExtraRepositories):
    """In memory task data for the Piuparts task."""

    input: PiupartsDataInput
    base_tgz: LookupSingle
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle


class PiupartsDynamicData(
    BaseDynamicTaskDataWithExecutor, BaseDynamicTaskDataWithExtraRepositories
):
    """Dynamic data for the Piuparts task."""

    # BaseTaskDynamicDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    input_binary_artifacts_ids: list[int]
    base_tgz_id: int

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [
            self.environment_id,
            self.base_tgz_id,
            *self.input_binary_artifacts_ids,
        ]


class SbuildInput(BaseTaskDataModel):
    """Input for a sbuild task."""

    source_artifact: LookupSingle
    extra_binary_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )


class SbuildBinNMU(BaseTaskDataModel):
    """binmu for a sbuild task."""

    # --make-binNMU
    changelog: str
    # --append-to-version
    suffix: str
    # --binNMU-timestamp, default to now
    timestamp: dt.datetime | None = None
    # --maintainer, defaults to uploader
    maintainer: pydantic.NameEmail | None = None


class SbuildBuildComponent(StrEnum):
    """Possible values for build_components."""

    ANY = "any"
    ALL = "all"
    SOURCE = "source"


class SbuildBuildDepResolver(StrEnum):
    """Possible values for build_dep_resolver."""

    APT = "apt"
    APTITUDE = "aptitude"
    ASPCUD = "aspcud"
    XAPT = "xapt"


class SbuildData(BaseTaskDataWithExecutor, BaseTaskDataWithExtraRepositories):
    """In memory task data for the Sbuild task."""

    input: SbuildInput
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    build_architecture: str
    build_components: list[SbuildBuildComponent] = pydantic.Field(
        default_factory=lambda: [SbuildBuildComponent.ANY]
    )
    binnmu: SbuildBinNMU | None = None
    build_profiles: list[str] | None = None
    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle
    build_dep_resolver: SbuildBuildDepResolver | None = None
    aspcud_criteria: str | None = None
    resolve_alternatives: bool = False

    @pydantic.model_validator(mode="after")
    def check_binnmu_against_components(self) -> Self:
        """Binnmus are incompatible with architecture-independent builds."""
        if self.binnmu is not None:
            if SbuildBuildComponent.ALL in self.build_components:
                raise ValueError(
                    "Cannot build architecture-independent packages in a binNMU"
                )
        return self


class SbuildDynamicData(
    BaseDynamicTaskDataWithExecutor, BaseDynamicTaskDataWithExtraRepositories
):
    """Dynamic data for the Sbuild task."""

    input_source_artifact_id: int
    input_extra_binary_artifacts_ids: list[int] = pydantic.Field(
        default_factory=list
    )
    binnmu_maintainer: str | None = None

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        result: list[int] = []
        if val := self.environment_id:
            result.append(val)
        result.append(self.input_source_artifact_id)
        result.extend(self.input_extra_binary_artifacts_ids)
        return result


class ImageCacheUsageLogEntry(BaseTaskDataModel):
    """Entry in ImageCacheUsageLog for cached executor images."""

    filename: str
    backend: BackendType | None = None
    timestamp: dt.datetime

    @pydantic.field_validator("timestamp", mode="after")
    @classmethod
    def timestamp_is_aware(cls, timestamp: dt.datetime) -> dt.datetime:
        """Ensure that the timestamp is TZ-aware."""
        tzinfo = timestamp.tzinfo
        if tzinfo is None or tzinfo.utcoffset(timestamp) is None:
            raise ValueError("timestamp is TZ-naive")
        return timestamp


class ImageCacheUsageLog(BaseTaskDataModel):
    """Usage log for cached executor images."""

    version: int = 1
    backends: set[BackendType] = pydantic.Field(default_factory=set)
    usage: list[ImageCacheUsageLogEntry] = pydantic.Field(default_factory=list)

    @pydantic.field_validator("version", mode="after")
    @classmethod
    def version_is_known(cls, version: int) -> int:
        """Ensure that the version is known."""
        if version != 1:
            raise ValueError(f"Unknown usage log version {version}")
        return version


class ExtractForSigningInput(BaseTaskDataModel):
    """Input for the ExtractForSigning task."""

    template_artifact: LookupSingle
    binary_artifacts: LookupMultiple


class ExtractForSigningData(BaseTaskDataWithExecutor):
    """In-memory task data for the ExtractForSigning task."""

    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle

    input: ExtractForSigningInput

    @pydantic.field_validator("backend", mode="after")
    @classmethod
    def backend_is_auto(cls, backend: str) -> str:
        """Ensure that the backend is "auto"."""
        if backend != BackendType.AUTO:
            raise ValueError(
                f'ExtractForSigning only accepts backend '
                f'"{BackendType.AUTO}", not "{backend}"'
            )
        return backend


class ExtractForSigningDynamicData(BaseDynamicTaskDataWithExecutor):
    """Dynamic data for the ExtractForSigning task."""

    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    input_template_artifact_id: int
    input_binary_artifacts_ids: list[int]

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [
            self.environment_id,
            self.input_template_artifact_id,
            *self.input_binary_artifacts_ids,
        ]


class AssembleSignedSourceData(BaseTaskDataWithExecutor):
    """In-memory task data for the AssembleSignedSource task."""

    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle

    template: LookupSingle
    signed: LookupMultiple

    @pydantic.field_validator("backend", mode="after")
    @classmethod
    def backend_is_auto(cls, backend: str) -> str:
        """Ensure that the backend is "auto"."""
        if backend != BackendType.AUTO:
            raise ValueError(
                f'AssembleSignedSource only accepts backend '
                f'"{BackendType.AUTO}", not "{backend}"'
            )
        return backend


class AssembleSignedSourceDynamicData(BaseDynamicTaskDataWithExecutor):
    """Dynamic data for the AssembleSignedSource task."""

    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    template_id: int
    signed_ids: list[int]
    unsigned_ids: list[int] = pydantic.Field(default_factory=list)

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [
            self.environment_id,
            self.template_id,
            *self.signed_ids,
            *self.unsigned_ids,
        ]


class MakeSourcePackageUploadInput(BaseTaskDataModel):
    """Input for a MakeSourcePackageUpload task."""

    source_artifact: LookupSingle


class MakeSourcePackageUploadData(BaseTaskDataWithExecutor):
    """In memory task data for the MakeSourcePackageUpload task."""

    # BaseTaskDataWithExecutor declares this as optional, but it's required
    # here.
    environment: LookupSingle

    input: MakeSourcePackageUploadInput
    since_version: str | None = None
    target_distribution: str | None = None


class MakeSourcePackageUploadDynamicData(BaseDynamicTaskDataWithExecutor):
    """Expanded dynamic data for the MakeSourcePackageUpload task."""

    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
    # required here.
    environment_id: int

    input_source_artifact_id: int

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.environment_id, self.input_source_artifact_id]


class MergeUploadsInput(BaseTaskDataModel):
    """Input for a MergeUploads task."""

    uploads: LookupMultiple


class MergeUploadsData(BaseTaskData):
    """In memory task data for the MergeUploads task."""

    # No longer used.
    backend: BackendType = BackendType.AUTO
    # No longer used.
    environment: LookupSingle | None = None

    input: MergeUploadsInput


class MergeUploadsDynamicData(BaseDynamicTaskData):
    """Expanded dynamic data for the MergeUploads task."""

    # No longer used.
    environment_id: int | None = None

    input_uploads_ids: list[int]

    @override
    def get_source_package_name(self) -> str | None:
        return self.subject

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return self.input_uploads_ids


class BaseTaskInput(pydantic.BaseModel):
    """Base for structures defining task input artifacts and collections."""

    model_config = pydantic.ConfigDict(
        validate_assignment=True, extra="forbid", frozen=True
    )


class BaseLabeledTaskInput(BaseTaskInput):
    """Task input annotated with a task data field name."""

    #: the task data label.
    label: str


class BaseTaskInputArtifact(BaseTaskInput):
    """Represent an artifact as task input."""

    #: Artifact ID
    artifact_id: int | None = None
    #: Artifact category
    category: ArtifactCategory | None = None
    #: Artifact data
    data: ArtifactData | None = None

    @pydantic.field_serializer("data")
    def serialize_data(
        self, data: ArtifactData | None
    ) -> dict[str, Any] | None:
        """Serialize the data field."""
        if data is None:
            return None
        return data.model_dump()

    @pydantic.field_validator("category", mode="before")
    @classmethod
    def _validate_category(cls, category: Any) -> Any:
        """Validate the category field."""
        if category is None:
            return None
        if category not in ARTIFACT_DATA_MODELS_BY_CATEGORY:
            raise ValueError(f"Category {category!r} is not supported")
        return category

    @pydantic.field_validator("data", mode="before")
    @classmethod
    def _validate_data(cls, data: Any, info: pydantic.ValidationInfo) -> Any:
        """Validate and deserialize the data field."""
        if data is None:
            return None
        if (category := info.data.get("category")) is None:
            raise ValueError("Category must be present if data is supplied")
        artifact_cls = ARTIFACT_DATA_MODELS_BY_CATEGORY[category]
        return artifact_cls.model_validate(data)


class BaseInputArtifactSingle(BaseTaskInputArtifact, BaseLabeledTaskInput):
    """Single task input artifact."""

    #: Lookup used to obtain the artifact. Usually from the task data but
    #: could be generated by the task logic.
    lookup: LookupSingle


class InputArtifactSingle(BaseInputArtifactSingle):
    """Represent a declared single input artifact for a task."""

    def resolved_ids(self) -> list[int]:
        """Return [artifact_id] or []."""
        return [] if self.artifact_id is None else [self.artifact_id]


class InputArtifactMultiple(BaseLabeledTaskInput):
    """Represent declared multiple input artifacts for a task."""

    #: Lookup used to obtain the artifacts.
    lookup: LookupMultiple

    #: Artifact data if it's already known
    artifacts: Sequence[BaseTaskInputArtifact] = pydantic.Field(
        default_factory=list
    )

    def resolved_ids(self) -> list[int]:
        """Return artifact_ids or []."""
        return [
            a.artifact_id for a in self.artifacts if a.artifact_id is not None
        ]

    @classmethod
    def from_ids(
        cls,
        lookup: LookupMultiple,
        label: str,
        artifact_ids: Collection[int] | None,
    ) -> Self:
        """Create an InputArtifactMultiple from a list of artifact IDs."""
        return cls(
            lookup=lookup,
            label=label,
            artifacts=(
                [
                    BaseTaskInputArtifact(artifact_id=artifact_id)
                    for artifact_id in artifact_ids or ()
                ]
            ),
        )


# TODO: serialize collection_id as id for shorter JSON in the database?
class InputCollectionSingle(BaseLabeledTaskInput):
    """Information about a collection."""

    #: Lookup used to obtain the collection
    lookup: LookupSingle

    #: Collection ID
    collection_id: int

    #: Name of the scope that contains the collection
    scope_name: str

    #: Name of the workspace that contains the collection
    workspace_name: str

    #: Category of the collection
    category: CollectionCategory

    #: Collection name
    name: str

    #: Collection data
    data: dict[str, Any]


InputArtifact: TypeAlias = InputArtifactSingle | InputArtifactMultiple



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Logic to authenticate a request using signon Providers."""

import functools
import logging
from collections.abc import Iterable, Sequence

import django.http
from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth import backends, load_backend
from django.contrib.auth.hashers import make_password
from django.core.exceptions import (
    ImproperlyConfigured,
    PermissionDenied,
    ValidationError,
)

from debusine.db.models import Group, Identity, User
from debusine.server.signon import providers
from debusine.server.signon.auth import SignonAuthBackend
from debusine.server.signon.signon_utils import split_full_name

log = logging.getLogger("debusine.server.signon")


class MapIdentityFailed(Exception):
    """Exception raised when unable to map an identity to a user."""


class Signon:
    """
    Backend used to interact with external authentication providers.

    This is setup by SignonMiddleware as request.signon.

    The constructor needs to be as lightweight as possible, as it is called on
    every request. Everything else is loaded only when needed.
    """

    def __init__(self, request: django.http.HttpRequest) -> None:
        """Create a Signon object for a request."""
        self.request = request
        self.providers: Sequence[providers.Provider] = getattr(
            settings, "SIGNON_PROVIDERS", ()
        )

    def status(
        self,
    ) -> Iterable[tuple[providers.BoundProvider, Identity | None]]:
        """
        Query the status of remote authentication providers.

        :returns: an iterable of ``(bound_provider, identity | None)``
        """
        for provider in self.providers:
            bound = provider.bind(self.request)
            identity = self.identities.get(provider.name)
            yield bound, identity

    @functools.cached_property
    def identities(self) -> dict[str, Identity]:
        """Lazily populate self.identities."""
        return self._compute_identities()

    def logout_identities(self) -> None:
        """Deactivate all active external identities."""
        for provider in self.providers:
            bound = provider.bind(self.request)
            bound.logout()
        self._remove_invalid_signon_user()

    def _compute_identities(self) -> dict[str, Identity]:
        """
        Instantiate valid Identity entries for this request.

        Delegate Provider objects with looking up valid Identity objects from
        the current request.
        """
        identities = {}

        for provider in self.providers:
            pk = self.request.session.get(f"signon_identity_{provider.name}")
            if pk is None:
                continue

            try:
                identity = Identity.objects.get(pk=pk, issuer=provider.name)
            except Identity.DoesNotExist:
                # If the session has a broken Identity ID, remove it
                del self.request.session[f"signon_identity_{provider.name}"]
                continue

            identities[provider.name] = identity

        return identities

    def validate_claims(
        self, provider: providers.Provider, identity: "Identity"
    ) -> list[str]:
        """
        Check if the claims in identity are acceptable.

        This is the function that decides whether to accept or reject a remote
        login. It provides a generic default implementation, and can be
        overridden by the class set in ``SIGNON_CLASS`` to implement more
        complex checks.

        By default, it honors the ``restrict`` configuration in ``GitLab``
        providers, for ``email_verified`` and ``group:<groupname>``. For
        example::

            SIGNON_PROVIDERS = [
                providers.ProviderClass(
                    # Restrict local user creation to users with these matching
                    # claims
                    restrict=("email-verified", "group:debian"),
                )
            ]

        :return: a list of error messages, or an empty list if validation passed
        """
        res = []
        if isinstance(provider, providers.OIDCProvider):
            for restriction in provider.restrict:
                if restriction == "email-verified":
                    # Work only with verified emails
                    if not identity.claims.get("email_verified", False):
                        res.append(
                            f"identity {identity}"
                            " does not have a verified email"
                        )
                elif restriction.startswith("group:"):
                    group_names = restriction[6:].split(",")
                    group_claims = identity.claims.get(
                        provider.group_claim_key, ()
                    )
                    for group_name in group_names:
                        if group_name in group_claims:
                            break
                    else:
                        if len(group_names) == 1:
                            group_list = repr(group_names[0])
                        else:
                            group_list = (
                                ", ".join(repr(x) for x in group_names[:-1])
                                + f" or {group_names[-1]!r}"
                            )
                        res.append(
                            f"identity {identity} is not in group {group_list}"
                        )
                else:
                    raise ImproperlyConfigured(
                        f"unsupported signon restriction: {restriction!r}"
                    )
        return res

    def identity_bind_to_current_user(self, identity: Identity) -> None:
        """Bind the identity to the current user."""
        # Checked by activate_identity
        assert self.request.user.is_authenticated
        log.info("%s: auto associated to %s", self.request.user, identity)
        identity.user = self.request.user
        identity.save()

    def identity_login_bound_user(self, identity: Identity) -> None:
        """Log in the user already bound to the identity."""
        # Checked by activate_identity
        assert identity.user is not None
        auth_backend_name = "debusine.server.signon.auth.SignonAuthBackend"
        backend = auth.load_backend(auth_backend_name)
        assert isinstance(backend, backends.ModelBackend)
        if not backend.user_can_authenticate(identity.user):
            raise PermissionDenied("user is not allowed to log in")

        log.debug("logging in user %s", identity.user)
        auth.login(
            self.request,
            identity.user,
            backend=auth_backend_name,
        )

    def identity_create_user(self, identity: Identity) -> None:
        """Attempt to create and bind a user for the identity."""
        # Validate identity claims
        if not (provider := self.get_provider_for_identity(identity)):
            raise MapIdentityFailed(
                f"Invalid provider {identity.issuer!r} in identity"
            )
        if claim_errors := self.validate_claims(provider, identity):
            for error in claim_errors:
                log.warning("%s", error)
            raise MapIdentityFailed(
                "Identity claims did not validate: " + "; ".join(claim_errors)
            )

        # First lookup an existing user
        user = self._lookup_user_from_identity(identity)
        if user is not None:
            log.info("%s: user matched to identity %s", user, identity)
        else:
            # Else try to create a new user
            user = self.create_user_from_identity(identity)
            if user is not None:
                log.info("%s: auto created from identity %s", user, identity)
                self.setup_new_user(user, identity)
            else:
                raise MapIdentityFailed(
                    f"Cannot create a local user for {identity}"
                )

        # A user was found or created: bind it to the identity
        identity.user = user
        identity.save()
        log.info("%s: bound to identity %s", user, identity)

    def activate_identity(self, identity: Identity, *options: str) -> bool:
        """
        Activate the given identity and update authentication accordingly.

        :return: True if activation succeeded; False if it failed but the
          calling view can reasonably continue without logging in.
        :raises PermissionDenied: if the user bound to the identity is not
          allowed to log in.
        """
        if self.request.user.is_authenticated:
            if "bind" not in options:
                raise PermissionDenied("user is already logged in")
            self.identity_bind_to_current_user(identity)
        elif identity.user is None:
            try:
                self.identity_create_user(identity)
            except MapIdentityFailed as exc:
                messages.add_message(
                    self.request, messages.ERROR, f"Login failed: {exc}"
                )
            else:
                self.identity_login_bound_user(identity)
        else:
            self.identity_login_bound_user(identity)

        self.request.session[f"signon_identity_{identity.issuer}"] = identity.pk
        return identity.user is not None

    def _remove_invalid_signon_user(self) -> None:
        """
        Log out an externally authenticated user.

        This is used to invalidate credentials in case a consistency check
        failed between active identities.

        Log out only happens if the user was authenticated via SignonMiddleware
        """
        try:
            stored_backend = load_backend(
                self.request.session.get(auth.BACKEND_SESSION_KEY, '')
            )
        except ImportError:
            # backend failed to load
            auth.logout(self.request)
        else:
            if isinstance(stored_backend, SignonAuthBackend):
                auth.logout(self.request)

    def get_provider_for_identity(
        self, identity: Identity
    ) -> providers.Provider | None:
        """Find the Provider for an identity."""
        for provider in self.providers:
            if provider.name == identity.issuer:
                return provider

        log.warning(
            "identity %s has unknown issuer %s", identity, identity.issuer
        )
        return None

    def _lookup_user_from_identity(self, identity: Identity) -> User | None:
        """Lookup an existing user from claims in an Identity."""
        User = auth.get_user_model()
        try:
            return User.objects.get(email=identity.claims["email"])
        except User.DoesNotExist:
            return None

    def create_user_from_identity(self, identity: Identity) -> User | None:
        """
        Create a user from the data in an Identity.

        This is the function that creates a local user after the first
        successful login with a remote identity, and no matching user is found
        that already exists.

        This is only called if the identity has passed ``validate_claims``
        checks, so those checks can be assumed as invariants.
        """
        User = auth.get_user_model()
        first_name, last_name = split_full_name(identity.claims["name"])

        # Django does not run validators on create_user, so garbage in the
        # claims can either create garbage users, or cause database transaction
        # errors that will invalidate the current transaction.
        #
        # See: https://stackoverflow.com/questions/67442439/why-django-does-not-validate-email-in-customuser-model  # noqa: E501

        # Instead of calling create_user, I instead have to replicate what it
        # does here and call validation explicitly before save.

        # This is the equivalent of the following, with validation:
        # user = User.objects.create_user(
        #     username=identity.claims["email"],
        #     first_name=first_name,
        #     last_name=last_name,
        #     email=identity.claims["email"],
        # )
        email = User.objects.normalize_email(identity.claims["email"])
        username = User.normalize_username(identity.claims["email"])
        user = User(
            username=username,
            email=email,
            first_name=first_name,
            last_name=last_name,
        )
        user.password = make_password(None)

        try:
            user.clean_fields()
        except ValidationError as e:
            log.warning("%s: cannot create a local user", identity, exc_info=e)
            return None

        user.save()
        return user

    def setup_new_user(self, user: User, identity: Identity) -> None:
        """
        Set up a newly created user.

        This is only called if the identity has passed ``validate_claims``
        checks, and a new user has just been created.

        By default it provides a generic implementation, and can be overridden
        by the class set in ``SIGNON_CLASS`` to implement more complex
        behaviours.

        The default implementation honors the ``add_to_group`` option set in
        the Provider instance. For example::

            SIGNON_PROVIDERS = [
                providers.ProviderClass(
                    options={
                        "add_to_group": {
                            # Map gitlab group names to Debusine groups
                            "gitlab/group": "scopename/groupname",
                            # Map nm.debian.org statuses to Debusine groups
                            "nm:dm": "scopename/groupname",
                            "nm:dm_ga": "scopename/groupname",
                        },
                    }
                )
            ]
        """
        if (provider := self.get_provider_for_identity(identity)) is None:
            return

        if not isinstance(provider, providers.OIDCProvider):
            return

        if (add_to_group := provider.options.get("add_to_group", None)) is None:
            return

        self.add_user_to_groups(user, identity, add_to_group)

    def add_user_to_groups(
        self, user: User, identity: Identity, add_to_group: dict[str, str]
    ) -> None:
        """Add a user to groups according to ``add_to_group`` rules."""
        provider = self.get_provider_for_identity(identity)
        # Already checked in setup_new_user
        assert provider is not None
        assert isinstance(provider, providers.OIDCProvider)

        claim_groups = identity.claims.get(provider.group_claim_key, ())
        for remote_group, debusine_group in add_to_group.items():
            if remote_group not in claim_groups:
                continue
            self.add_user_to_group(user, debusine_group)

    def add_user_to_group(self, user: User, scoped_name: str) -> None:
        """Add a user to the named ``scope/[workspace/]group`` group."""
        try:
            group = Group.objects.get_by_name(scoped_name)
        except Group.DoesNotExist:
            raise ImproperlyConfigured(f"Group {scoped_name!r} not found")
        group.users.add(user)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebDiff work request plugin."""

import textwrap
from typing import ClassVar
from unittest import mock

from django.template.response import SimpleTemplateResponse
from rest_framework import status
from unidiff.patch import PatchedFile

from debusine.artifacts.models import (
    ArtifactCategory,
    ArtifactData,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
)
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    WorkRequest,
    Workspace,
)
from debusine.tasks import DebDiff
from debusine.tasks.models import (
    DebDiffData,
    DebDiffDynamicData,
    DebDiffFlags,
    DebDiffInput,
    LookupMultiple,
)
from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin


class DebDiffViewArtifactPluginTests(ViewTestMixin, TestCase):
    """Tests for DebDiffViewArtifactPlugin."""

    workspace: ClassVar[Workspace]
    work_request: ClassVar[WorkRequest]
    debdiff_artifact: ClassVar[Artifact]

    @classmethod
    def create_artifact_with_relates(
        cls, artifact: Artifact, target_artifact_category: ArtifactCategory
    ) -> None:
        data: ArtifactData

        match target_artifact_category:
            case ArtifactCategory.UPLOAD:
                data = DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "amd64",
                        "Files": [{"name": "hello_1.0_amd64.deb"}],
                    },
                )
            case ArtifactCategory.BINARY_PACKAGE:
                data = DebianBinaryPackage(
                    srcpkg_name="hello_1.0_amd64.deb",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "hello",
                        "Version": "1.0",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                )
            case ArtifactCategory.SOURCE_PACKAGE:
                data = DebianSourcePackage(
                    name="hello",
                    version="1.0",
                    type="dpkg",
                    dsc_fields={},
                )
            case _ as unreachable:
                raise NotImplementedError(f"{unreachable!r} not supported")

        target, _ = cls.playground.create_artifact(
            category=target_artifact_category, data=data
        )

        ArtifactRelation.objects.create(
            artifact=artifact,
            target=target,
            type=ArtifactRelation.Relations.RELATES_TO,
        )

    def test_button_content(self) -> None:
        """Test DebDiff button found and debdiff contents is included."""
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: b""},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        debdiff_a = self.assertHasElement(tree, "//a[@id='nav-debdiff-tab']")
        self.assertTextContentEqual(debdiff_a, "DebDiff")

        debdiff_contents = self.assertHasElement(
            tree, "//div[@id='nav-debdiff']"
        )
        h2 = debdiff_contents.xpath(".//h2")

        self.assertEqual(h2, ["Request"])

        # Assert content is displayed by default
        self.assertIn("show active", debdiff_contents.attrib["class"])

        files_contents = self.assertHasElement(tree, "//div[@id='nav-files']")
        self.assertNotIn("show active", files_contents.attrib["class"])

    def test_binary_artifacts_identical(self) -> None:
        debdiff_report_identical = textwrap.dedent(
            """
            File lists identical (after any substitutions)

            No differences were encountered between the control files
            """
        ).encode("utf-8")
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report_identical},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert original and new files
        original = self.assertHasElement(tree, "//li[@id='original']")
        self.assertTextContentEqual(original, "Original: hello-1.deb")

        new = self.assertHasElement(tree, "//li[@id='new']")
        self.assertTextContentEqual(new, "New: hello-2.deb")

        # Assert text no changes reported
        no_changes_reported = self.assertHasElement(
            tree, "//p[@id='debdiff-report-no-differences']"
        )
        self.assertTextContentEqual(
            no_changes_reported, "DebDiff did not report any differences."
        )

    def test_binary_cannot_parse_too_short(self) -> None:
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: b""},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-parsing-debdiff-output']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Error parsing debdiff output: "
            "Cannot parse any information from debdiff output ",
        )

    def test_binary_cannot_parse_line(self) -> None:
        artifact, _ = self.playground.create_artifact(
            paths={
                DebDiff.CAPTURE_OUTPUT_FILENAME: b"Some line\n"
                b"\n"
                b"Files in second .changes but not in first\n"
                b"\n"
                b"Unrecognised\tline\n"
            },
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-parsing-debdiff-output']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Error parsing debdiff output: "
            "Failed to parse line: 'Unrecognised\\tline'",
        )

    def test_binary_all_sections_with_changes(self) -> None:
        debdiff_report = textwrap.dedent(
            """
            some debdiff text header"

            Files in second .deb but not in first
            -------------------------------------
            -rw-r--r--  root/root   /etc/simplemonitor/monitor.ini
            -rw-r--r--  root/root   DEBIAN/conffiles

            Files in first .deb but not in second
            ------------------------------------
            -rw-r--r--  root/root   /usr/lib/python3/dist-packages/pyaarlo
            -rw-r--r--  root/root   /usr/share/doc/python3-pyaarlo/READ.gz

            Control files: lines which differ (wdiff format)
            ------------------------------------------------
            Description: [-one description-] {+another description+}
            Homepage: [-https://github.com/twrecked/pyaarlo-]
            {+https://simplemonitor.readthedocs.io+}"""
        ).encode("utf-8")
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert files in original not in new
        files_in_original_not_in_new = self.assertHasElement(
            tree, "//h4[@id='files-in-original-not-in-new']"
        )
        files = self.assertHasElement(
            files_in_original_not_in_new, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            files,
            "-rw-r--r--  root/root   /usr/lib/python3/dist-packages/pyaarlo\n"
            "-rw-r--r--  root/root   /usr/share/doc/python3-pyaarlo/READ.gz\n",
        )

        # Assert files in new not in original
        files_in_new_not_in_original = self.assertHasElement(
            tree, "//h4[@id='files-in-new-not-in-original']"
        )
        files = self.assertHasElement(
            files_in_new_not_in_original, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            files,
            "-rw-r--r--  root/root   /etc/simplemonitor/monitor.ini\n"
            "-rw-r--r--  root/root   DEBIAN/conffiles\n",
        )

        # Assert control diffs changes
        control_diffs_changes = self.assertHasElement(
            tree, "//h4[@id='control-diffs-changes']"
        )
        files = self.assertHasElement(
            control_diffs_changes, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            files,
            "Description: [-one description-] {+another description+}\n"
            "Homepage: [-https://github.com/twrecked/pyaarlo-] "
            "{+https://simplemonitor.readthedocs.io+}\n",
        )

    def test_binary_all_sections_with_no_control_package_changes(self) -> None:
        debdiff_report = textwrap.dedent(
            """
            File lists identical (after any substitutions)

            No differences were encountered between the control files of package python3-ping3

            No differences were encountered between the control files of package simplemonitor
            """  # noqa: E501
        ).encode("utf-8")

        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={
                "original": "python3-ping-1.deb simplemonitor-1.deb",
                "new": "python3-ping-2.deb simplemonitor-2.deb",
            },
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert control diffs changes
        control_diffs_changes = self.assertHasElement(
            tree, "//h4[@id='control-diffs-changes']"
        )
        text = self.assertHasElement(
            control_diffs_changes, "following-sibling::p[1]"
        )
        self.assertTextContentEqual(
            text,
            "No differences were encountered "
            "between the control files of packages:",
        )

        ul = self.assertHasElement(
            control_diffs_changes, "following-sibling::ul[1]"
        )

        li_elements = ul.xpath("./li")

        self.assertEqual(
            [li.text.strip() for li in li_elements],
            ["python3-ping3", "simplemonitor"],
        )

    def test_binary_control_package_changes_per_package(self) -> None:
        debdiff_report = textwrap.dedent(
            """
            [The following lists of changes regard files as different if they have
            different names, permissions or owners.]

            Control files of package python3-ping3: lines which differ (wdiff format)
            -------------------------------------------------------------------------
            Installed-Size: [-100-] {+101+}
            Version: [-4.0.4-2-] {+4.0.4-3+}

            Control files of package simplemonitor: lines which differ (wdiff format)
            -------------------------------------------------------------------------
            Installed-Size: [-2013-] {+2043+}
            Version: [-1.12.1-1-] {+1.13.0-1+}
            """  # noqa: E501
        ).encode("utf-8")
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={
                "original": "python3-ping-1.deb simplemonitor-1.deb",
                "new": "python3-ping-2.deb simplemonitor-2.deb",
            },
        )
        self.create_artifact_with_relates(artifact, ArtifactCategory.UPLOAD)

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert control diffs changes
        control_diffs_changes = self.assertHasElement(
            tree, "//h4[@id='control-diffs-changes']"
        )

        # check python3-ping3
        python3_ping3_title = self.assertHasElement(
            control_diffs_changes, "following-sibling::h5[1]"
        )
        self.assertTextContentEqual(
            python3_ping3_title, "Package: python3-ping3"
        )
        python3_ping3_change = self.assertHasElement(
            python3_ping3_title, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            python3_ping3_change,
            "Installed-Size: [-100-] {+101+}\n"
            "Version: [-4.0.4-2-] {+4.0.4-3+}\n",
        )

        # check simplemonitor
        simplemonitor_title = self.assertHasElement(
            control_diffs_changes, "following-sibling::h5[2]"
        )
        self.assertTextContentEqual(
            simplemonitor_title, "Package: simplemonitor"
        )
        simplemonitor_change = self.assertHasElement(
            simplemonitor_title, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            simplemonitor_change,
            "Installed-Size: [-2013-] {+2043+}\n"
            "Version: [-1.12.1-1-] {+1.13.0-1+}\n",
        )

    def test_binary_set_of_debs(self) -> None:
        debdiff_report = textwrap.dedent(
            """\
            Same package name appears more than once (possibly due to renaming): simplemonitor
            [The following lists of changes regard files as different if they have
            different names, permissions or owners.]

            Files in second set of .debs but not in first
            ---------------------------------------------
            -rw-r--r--  root/root   /lib/s/system/simplemonitor.service
            -rw-r--r--  root/root   /simplemonitor-1.1.0.dist-info/METADATA

            Files in first set of .debs but not in second
            ---------------------------------------------
            -rw-r--r--  root/root   .../s.../INSTALLER

            Control files of package simplemonitor: lines which differ (wdiff format)
            -------------------------------------------------------------------------
            Depends: adduser, python3-arrow, python3-boto3, python3-colorlog, [-python3-icmplib,-] python3-importlib-metadata, python3-jinja2, python3-paho-mqtt, python3-paramiko, python3-ping3, python3-psutil, python3-pyaarlo, python3-requests, python3-ring-doorbell, python3-twilio, python3:any, [-libjs-jquery (>= 3.6.0),-] libjs-sphinxdoc (>= [-8.1),-] {+5.2),+} sphinx-rtd-theme-common (>= [-3.0.2+dfsg)-] {+1.2.0+dfsg)+}
             [-.-]
            [- The package include a systemd service.-]
            Installed-Size: [-2298-] {+2006+}
            [-Recommends: systemd-]
            Version: [-1.14.0a-3-] {+1.12.0-1+}
            """  # noqa: E501
        ).encode("utf-8")

        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={
                "original": (
                    "python3-ping3_4.0.4-2_all.deb "
                    "simplemonitor_1.12.1-1_all.deb"
                ),
                "new": (
                    "python3-ping3_4.0.4-3_all.deb "
                    "simplemonitor_1.13.0-1_all.deb"
                ),
            },
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.BINARY_PACKAGE
        )

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert files in original not in new
        files_in_original_not_in_new = self.assertHasElement(
            tree, "//h4[@id='files-in-original-not-in-new']"
        )
        files = self.assertHasElement(
            files_in_original_not_in_new, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            files,
            "-rw-r--r--  root/root   .../s.../INSTALLER\n",
        )

        files_in_new_not_in_original = self.assertHasElement(
            tree, "//h4[@id='files-in-new-not-in-original']"
        )
        files = self.assertHasElement(
            files_in_new_not_in_original, "following-sibling::pre[1]"
        )
        self.assertTextContentEqual(
            files,
            "-rw-r--r--  root/root   /lib/s/system/simplemonitor.service\n"
            "-rw-r--r--  root/root   /simplemonitor-1.1.0.dist-info/METADATA\n",
        )

    def test_source(self) -> None:
        debdiff_report = textwrap.dedent(
            """\
            diff -u -N original/added.txt new/added.txt
            --- original/added.txt	1970-01-01 01:00:00.000000000 +0100
            +++ new/added.txt	2025-03-26 12:42:27.672906377 +0000
            @@ -0,0 +1 @@
            +new file
            diff -u -N original/changed.txt new/changed.txt
            --- original/changed.txt	2025-03-26 12:41:44.191924503 +0000
            +++ new/changed.txt	2025-03-26 12:42:21.945798421 +0000
            @@ -1,3 +1,2 @@
             1
             2
            -3
            diff -u -N original/removed.txt new/removed.txt
            --- original/removed.txt	2025-03-26 12:42:37.607381416 +0000
            +++ new/removed.txt	1970-01-01 01:00:00.000000000 +0100
            @@ -1 +0,0 @@
            -removed file
        """
        ).encode("utf-8")
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.SOURCE_PACKAGE
        )
        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert files basic templating
        summary_changes = self.assertHasElement(
            tree, "//h2[@id='source-summary-of-changes']"
        )
        files = self.assertHasElement(
            summary_changes, "following-sibling::ul[1]"
        )

        self.assertTextContentEqual(
            files, "new/added.txt new/changed.txt new/removed.txt"
        )

        # Assert context data is correct
        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        context = response.context_data
        self.assertEqual(
            context["debdiff_source_summary"],
            [
                {
                    "diff_line_number": 2,
                    "operation": "added",
                    "path": "new/added.txt",
                },
                {
                    "diff_line_number": 7,
                    "operation": "modified",
                    "path": "new/changed.txt",
                },
                {
                    "diff_line_number": 14,
                    "operation": "removed",
                    "path": "new/removed.txt",
                },
            ],
        )

        # Context has the debdiff_file_tag
        self.assertEqual(context["debdiff_file_tag"], "debdiff")

        # Links to the lines are done correctly
        self.assertEqual(files.li[0].a.attrib["href"], "#L2-debdiff")
        self.assertEqual(files.li[1].a.attrib["href"], "#L7-debdiff")
        self.assertEqual(files.li[2].a.attrib["href"], "#L14-debdiff")

    def test_source_with_binary_files(self) -> None:
        debdiff_report = textwrap.dedent(
            """\
            diff -u -N original/added.txt new/added.txt
            --- original/added.txt	1970-01-01 01:00:00.000000000 +0100
            +++ new/added.txt	2025-03-26 12:42:27.672906377 +0000
            @@ -0,0 +1 @@
            +new file
            Binary files Binary files /t/p1/a.png and /t/p2/a.png differ
            Binary files Binary files /t/p1/b.png and /t/p2/b.png differ
        """
        ).encode("utf-8")
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.SOURCE_PACKAGE
        )
        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert files basic templating
        summary_changes = self.assertHasElement(
            tree, "//h2[@id='source-summary-of-changes']"
        )
        files = self.assertHasElement(
            summary_changes, "following-sibling::ul[1]"
        )

        self.assertTextContentEqual(
            files, "new/added.txt /t/p2/a.png /t/p2/b.png"
        )

        # Assert context data is correct
        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        self.assertEqual(
            response.context_data["debdiff_source_summary"],
            [
                {
                    "diff_line_number": 2,
                    "operation": "added",
                    "path": "new/added.txt",
                },
                {
                    "diff_line_number": 6,
                    "operation": "modified",
                    "path": "/t/p2/a.png",
                },
                {
                    "diff_line_number": 7,
                    "operation": "modified",
                    "path": "/t/p2/b.png",
                },
            ],
        )

    def test_source_unexpected_file_status(self) -> None:
        debdiff_report = textwrap.dedent(
            """\
            --- a.txt	2025-03-27 08:38:25.845875929 +0000
            +++ b.txt	2025-03-27 08:38:43.129990008 +0000
            @@ -1 +1 @@
            -1
            +2
        """
        ).encode("utf-8")

        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.SOURCE_PACKAGE
        )

        mock_file = mock.create_autospec(PatchedFile, instance=True)
        mock_file.is_added_file = False
        mock_file.is_removed_file = False
        mock_file.is_modified_file = False
        mock_file.path = "a.txt"

        with mock.patch(
            "debusine.web.views.debdiff.PatchSet", return_value=[mock_file]
        ):
            response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-parsing-debdiff-output']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Error parsing debdiff output: "
            "Unexpected file status in diff: a.txt",
        )

    def test_source_artifacts_identical(self) -> None:
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: b""},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.SOURCE_PACKAGE
        )

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        # Assert original and new as expected
        original = self.assertHasElement(tree, "//li[@id='original']")
        self.assertTextContentEqual(original, "Original: hello-1.dsc")
        new = self.assertHasElement(tree, "//li[@id='new']")
        self.assertTextContentEqual(new, "New: hello-2.dsc")

        # Assert text no changes reported
        no_changes_reported = self.assertHasElement(
            tree, "//p[@id='debdiff-report-no-differences']"
        )
        self.assertTextContentEqual(
            no_changes_reported, "DebDiff did not report any differences."
        )

    def test_get_context_data_related_to_renders_warning(self) -> None:
        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: b""},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )
        self.create_artifact_with_relates(
            artifact, ArtifactCategory.SOURCE_PACKAGE
        )
        artifact.relations.all().delete()

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-artifact']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Cannot display debdiff: "
            "Artifact is not complete (missing 'relates-to' relation)",
        )

    def test_get_context_data_no_debdiff_txt(self) -> None:
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-artifact']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Cannot display debdiff: Artifact does not contain debdiff.txt",
        )

        # Assert text report no differences
        self.assertFalse(tree.xpath("//p[@id='debdiff-report-no-differences']"))

    def test_get_context_data_file_in_artifact_error(self) -> None:
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.DEBDIFF,
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: b""},
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
            skip_add_files_in_store=True,
        )

        response = self.client.get(artifact.get_absolute_url())
        tree = self.assertResponseHTML(response)

        error = self.assertHasElement(
            tree,
            "//div[@id='error-artifact']",
        )

        self.assertTextContentEqual(
            error,
            'Cannot display debdiff: '
            'File "debdiff.txt" has not completed uploading.',
        )


class DebDiffViewWorkRequestPluginTests(TestCase):
    """Tests for DebDiffViewWorkRequestPlugin."""

    SAMPLE_TASK_DATA = {
        "input": {"source_artifacts": [421, 123]},
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
        "extra_flags": ["--dirs", "--nocontrol"],
    }

    def test_get_context_source_artifacts(self) -> None:
        source_artifact_original = (
            self.playground.create_minimal_source_package_artifact()
        )
        source_artifact_new = (
            self.playground.create_minimal_source_package_artifact()
        )
        environment = self.playground.create_debian_environment().artifact
        assert environment is not None

        work_request = self.playground.create_worker_task(
            task_name="debdiff",
            task_data=DebDiffData(
                input=DebDiffInput(
                    source_artifacts=[
                        source_artifact_original.id,
                        source_artifact_new.id,
                    ]
                ),
                build_architecture="amd64",
                environment=environment.id,
                extra_flags=[DebDiffFlags.DIRS, DebDiffFlags.NOCONTROL],
            ),
        )

        work_request.dynamic_task_data = DebDiffDynamicData(
            environment_id=environment.id,
            input_source_artifacts_ids=[
                source_artifact_original.id,
                source_artifact_new.id,
            ],
        ).model_dump()
        work_request.save()

        response = self.client.get(work_request.get_absolute_url())

        # Check context
        ctx = response.context

        description_data, description_template = (
            ctx["description_data"],
            ctx["description_template"],
        )

        self.assertEqual(
            description_data,
            {
                "environment_id": environment.id,
                "extra_arguments": ["--dirs", "--nocontrol"],
                "source_artifact_new_id": source_artifact_new.id,
                "source_artifact_original_id": source_artifact_original.id,
            },
        )

        self.assertEqual(description_template, "web/_debdiff-description.html")

        tree = self.assertResponseHTML(response)

        source_artifacts = self.assertHasElement(
            tree, "//p[@id='description-source_artifacts']"
        )
        self.assertTextContentEqual(
            source_artifacts,
            "Compare source package hello_1.0-1 with hello_1.0-1 using debdiff",
        )

        self.assertFalse(tree.xpath("//p[@id='description-binary_artifacts']"))

        environment_el = self.assertHasElement(
            tree, "//li[@id='description-executed_in_environment']"
        )
        self.assertTextContentEqual(
            environment_el, "Executed in environment test"
        )

        extra_arguments = self.assertHasElement(
            tree, "//li[@id='description-extra_arguments']"
        )
        self.assertTextContentEqual(
            extra_arguments, "Extra arguments: --dirs --nocontrol"
        )

    def test_get_context_binary_artifacts(self) -> None:
        binary_artifact_original_1 = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="bye", srcpkg_version="2.0-1"
            )
        )
        binary_artifact_original_2 = (
            self.playground.create_minimal_binary_package_artifact()
        )

        binary_artifact_new_1 = (
            self.playground.create_minimal_binary_package_artifact()
        )
        binary_artifact_new_2 = (
            self.playground.create_minimal_binary_package_artifact()
        )

        environment = self.playground.create_debian_environment().artifact
        assert environment is not None

        work_request = self.playground.create_worker_task(
            task_name="debdiff",
            task_data=DebDiffData(
                input=DebDiffInput(
                    binary_artifacts=[
                        LookupMultiple(
                            (
                                binary_artifact_original_1.id,
                                binary_artifact_original_2.id,
                            )
                        ),
                        LookupMultiple(
                            (binary_artifact_new_1.id, binary_artifact_new_2.id)
                        ),
                    ]
                ),
                build_architecture="amd64",
                environment=environment.id,
                extra_flags=[DebDiffFlags.DIRS, DebDiffFlags.NOCONTROL],
            ),
        )

        work_request.dynamic_task_data = DebDiffDynamicData(
            environment_id=environment.id,
            input_binary_artifacts_ids=[
                [binary_artifact_original_1.id, binary_artifact_original_2.id],
                [binary_artifact_new_1.id, binary_artifact_new_2.id],
            ],
        ).model_dump()
        work_request.save()

        response = self.client.get(work_request.get_absolute_url())

        # Check context
        ctx = response.context

        description_data, description_template = (
            ctx["description_data"],
            ctx["description_template"],
        )

        self.assertEqual(
            description_data,
            {
                "environment_id": environment.id,
                "extra_arguments": ["--dirs", "--nocontrol"],
                "binary_artifacts_new_count": 2,
                "source_package_names": [
                    "bye_2.0-1",
                    "hello_1.0-1",
                ],
                "binary_artifacts_architecture": "amd64",
            },
        )

        self.assertEqual(description_template, "web/_debdiff-description.html")

        tree = self.assertResponseHTML(response)

        binary_artifacts_el = self.assertHasElement(
            tree, "//p[@id='description-binary_artifacts']"
        )
        self.assertTextContentEqual(
            binary_artifacts_el,
            "Compare amd64 binaries from source packages "
            "bye_2.0-1, hello_1.0-1 using debdiff.",
        )

        self.assertFalse(tree.xpath("//p[@id='description-source_artifacts']"))

        environment_el = self.assertHasElement(
            tree, "//li[@id='description-executed_in_environment']"
        )
        self.assertTextContentEqual(
            environment_el, "Executed in environment test"
        )

        extra_arguments = self.assertHasElement(
            tree, "//li[@id='description-extra_arguments']"
        )
        self.assertTextContentEqual(
            extra_arguments, "Extra arguments: --dirs --nocontrol"
        )

    def test_get_context_data_description_without_packages(self) -> None:
        """'dynamic_data' does not contain either source or binary packages."""
        work_request = self.playground.create_worker_task(
            task_name="debdiff", task_data=self.SAMPLE_TASK_DATA
        )
        environment = self.playground.create_debian_environment().artifact
        assert environment is not None

        work_request.dynamic_task_data = DebDiffDynamicData(
            environment_id=environment.id,
            subject="hello",
            parameter_summary="hello: 1.0.0-1 → 1.0.0-2",
        ).model_dump()
        work_request.save()

        response = self.client.get(work_request.get_absolute_url())

        tree = self.assertResponseHTML(response)

        el = self.assertHasElement(tree, "//p[@id='general-description']")

        self.assertTextContentEqual(el, "Debdiff for hello: 1.0.0-1 → 1.0.0-2")

    def test_get_context_data_description_data_empty(self) -> None:
        """'description_data' is {}: 'dynamic_data' is None."""
        work_request = self.playground.create_worker_task(
            task_name="debdiff", task_data=self.SAMPLE_TASK_DATA
        )
        self.assertIsNone(work_request.dynamic_task_data)

        response = self.client.get(work_request.get_absolute_url())

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.context["description_data"], {})



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag library for users."""

from django import template
from django.utils.html import format_html

from debusine.db.context import context
from debusine.db.models import User

register = template.Library()


@register.filter
def user_link(user: User) -> str:
    """Return a link in an authenticated request, otherwise the name."""
    if user.can_display(context.pc):
        return format_html('<a href="{}">{}</a>', user.get_absolute_url(), user)
    else:
        return str(user)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Debusine exceptions."""

from typing import Any
from unittest import TestCase

from debusine.client.exceptions import DebusineError


class DebusineErrorTests(TestCase):
    """Tests for DebusineError exception."""

    def test_asdict(self) -> None:
        """Test asdict() returns the dictionary."""
        data: dict[str, Any] = {
            "title": "The title of the error",
            "detail": "The detail of the error",
            "status_code": 418,
        }
        debusine_error = DebusineError(**data)

        self.assertEqual(debusine_error.asdict(), data)

    def test_str(self) -> None:
        for exc, expected in (
            (DebusineError(title="title"), "title"),
            (DebusineError(title="title", status_code=418), "418 title"),
            (DebusineError(title="title", detail="desc"), "title: desc"),
            (
                DebusineError(title="title", detail="desc", status_code=418),
                "418 title: desc",
            ),
        ):
            with self.subTest(exc=exc.asdict()):
                self.assertEqual(str(exc), expected)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the DebusineHttpClient."""

import json
import re
from collections import deque
from io import BytesIO
from itertools import count
from typing import Any, override
from unittest import mock
from unittest.mock import ANY

import requests
import responses
from urllib3 import HTTPHeaderDict, HTTPResponse

from debusine.client import exceptions
from debusine.client.debusine_http_client import DebusineHttpClient
from debusine.client.exceptions import UnexpectedResponseError
from debusine.client.models import StrictBaseModel
from debusine.test import TestCase

try:  # pragma: no cover
    from responses.registries import OrderedRegistry
except ImportError:
    OrderedRegistry = None  # type: ignore[assignment,misc]


class SampleResponse(StrictBaseModel):
    """A dummy model for test cases."""

    id: int
    page: int | None = None


class DebusineHttpClientTests(TestCase):
    """Tests for the DebusineHttpClient."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.api_url = "https://debusine.debian.org/api/1.0"
        self.token = "token-for-server"

    def responses_add_test_endpoint(
        self,
        status: int,
        method: str = responses.GET,
        *,
        rsps: responses.RequestsMock = responses.mock,
    ) -> None:
        """
        Add a test response.

        Used for testing the debusine API client low level methods.
        """
        json = None
        if status == requests.codes.ok:  # pragma: no cover
            json = {'id': 12}
        rsps.add(
            method=method,
            url=self.api_url,
            json=json,
            status=status,
        )

    @staticmethod
    def responses_add_problem(
        method: str,
        url: str,
        title: str,
        detail: str | None = None,
        validation_errors: dict[str, str] | None = None,
        status: int = requests.codes.bad_request,
    ) -> None:
        """Add a debusine problem+json to responses."""
        response = {
            "title": title,
            "detail": detail,
            "validation_errors": validation_errors,
        }

        responses.add(
            method,
            url,
            json=response,
            content_type="application/problem+json",
            status=status,
        )

    @staticmethod
    def responses_add_paginated_responses(
        base_url: str,
        *,
        length: int = 4,
        page_size: int = 3,
        etag: str | None = None,
    ) -> list[SampleResponse]:
        """Add a set of responses for paginated listing requests."""
        all_objects = [(SampleResponse(id=i)) for i in range(length)]
        objects = deque(all_objects)
        extra_headers = {}

        if etag is not None:
            extra_headers["ETag"] = etag

        url = base_url
        for pageno in count(1):
            body: dict[str, Any] = {
                "count": length,
                "results": [],
            }
            for _ in range(page_size):
                if not objects:
                    break
                obj = objects.popleft()
                obj.page = pageno
                body["results"].append(obj.model_dump())
            if objects:
                body["next"] = base_url + f"?page={pageno + 1}"
            if pageno > 1:
                body["previous"] = base_url + f"?page={pageno - 1}"

            from responses.matchers import query_string_matcher

            url_without_query_string, _, query_string = url.partition("?")

            responses.add(
                responses.GET,
                url_without_query_string,
                json=body,
                match=[query_string_matcher(query_string)],
                adding_headers=extra_headers,
            )

            if not objects:
                break
            url = body["next"]
        else:  # pragma: no cover
            pass

        return all_objects

    def test_api_request_get_with_body_raise_value_error(self) -> None:
        """Raise ValueError if passing a body for the GET method."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            ValueError, "^data argument not allowed with HTTP GET$"
        ):
            debusine_client._api_request(
                "GET", self.api_url, SampleResponse, data={}
            )

    @responses.activate
    def test_api_request_expected_class_none_response_content_raise_error(
        self,
    ) -> None:
        """Raise UnexpectedResponseError: response was not empty."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)

        method = "PUT"
        content = "test"

        responses.add(method, self.api_url, body=content)

        expected_regexp = (
            rf"^Server \({self.api_url}\) expected to return "
            f"an empty body. Returned:\nb'{content}'$"
        )

        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError, expected_regexp
        ):
            debusine_client._api_request(method, self.api_url, None, data=None)

    def test_api_request_only_valid_methods(self) -> None:
        """Raise ValueError if the method is not recognized."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            ValueError, "^Method must be one of: GET, POST, PUT, PATCH, DELETE$"
        ):
            debusine_client._api_request("BAD-METHOD", "", SampleResponse)

    @responses.activate
    def test_api_request_not_found_raise_exception(self) -> None:
        """Raise NotFoundError if the request returns 404."""
        self.responses_add_test_endpoint(status=requests.codes.not_found)
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaises(exceptions.NotFoundError):
            debusine_client._api_request("GET", self.api_url, requests.codes.ok)

        self.assert_token_key_included_in_all_requests(self.token)

    @responses.activate
    def test_api_request_raise_debusine_error(self) -> None:
        """Raise DebusineError: server returned HTTP 400 and error message."""
        title = "Cannot update task"
        detail = "Invalid task_name"
        errors = {"field": "invalid value"}

        self.responses_add_problem(
            responses.GET, self.api_url, title, detail, errors
        )

        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaises(exceptions.DebusineError) as debusine_error:
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        exception = debusine_error.exception

        self.assertEqual(
            exception.asdict(),
            {"title": title, "detail": detail, "validation_errors": errors},
        )

    @responses.activate
    def test_api_request_raise_token_disabled_error(self) -> None:
        """Raise TokenDisabledError: server returned HTTP 403."""
        self.responses_add_test_endpoint(status=requests.codes.forbidden)

        token = "b3ecf243c43"
        debusine_client = DebusineHttpClient(self.api_url, token)
        with self.assertRaisesRegex(
            exceptions.ClientForbiddenError,
            rf"^HTTP 403. Token \({token}\) is invalid or disabled$",
        ):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

    @responses.activate
    def test_api_request_raise_debusine_error_404(self) -> None:
        """Raise DebusineError even with response.code == 404."""
        title = "Error"
        detail = "Workflow template not found"

        self.responses_add_problem(
            responses.POST,
            self.api_url,
            title,
            detail=detail,
            status=requests.codes.not_found,
        )
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaises(exceptions.DebusineError) as debusine_error:
            debusine_client._api_request("POST", self.api_url, SampleResponse)

        exception = debusine_error.exception

        self.assertEqual(exception.asdict(), {"title": title, "detail": detail})

    @responses.activate
    def test_api_request_raise_403_error_with_detail(self) -> None:
        """Raise TokenDisabledError: server returned HTTP 403 with detail."""
        title = "Error"
        detail = "You do not have permission to perform this action."

        self.responses_add_problem(
            responses.GET,
            self.api_url,
            title,
            detail=detail,
            status=requests.codes.forbidden,
        )

        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaises(exceptions.DebusineError) as debusine_error:
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        exception = debusine_error.exception

        self.assertEqual(
            exception.asdict(),
            {"title": title, "detail": detail},
        )

    @responses.activate
    def test_api_request_raise_unexpected_error(self) -> None:
        """Raise UnexpectedResponseError for unexpected status code."""
        self.responses_add_test_endpoint(status=requests.codes.teapot)
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            r"^Server returned unexpected status code: 418 "
            rf"\({self.api_url}\)$",
        ):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        self.assert_token_key_included_in_all_requests(self.token)

    @responses.activate
    def test_api_request_raise_unexpected_response_400_no_error_msg(
        self,
    ) -> None:
        """Raise UnexpectedError: server returned HTTP 400 and no error msg."""
        self.responses_add_test_endpoint(status=requests.codes.bad_request)
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            "^Server returned unexpected status code: 400 "
            rf"\({self.api_url}\)$",
        ):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

    @responses.activate
    def test_api_request_raise_unexpected_response_400_no_json(self) -> None:
        """Raise UnexpectedError: server returned HTTP 400 and no JSON."""
        responses.add(
            responses.GET,
            self.api_url,
            body="Not including JSON",
            status=requests.codes.bad_request,
        )
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            "^Server returned unexpected status code: 400 "
            rf"\({self.api_url}\)$",
        ):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

    def test_api_request_retry_errors(self) -> None:  # pragma: no cover
        """Some HTTP error responses are retried."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)

        for code in (
            requests.codes.server_error,  # 500
            requests.codes.service_unavailable,  # 503
            requests.codes.gateway_timeout,  # 504
            requests.codes.too_many_requests,  # 429
        ):
            for method in (
                responses.GET,
                responses.PATCH,
                responses.POST,
                responses.PUT,
            ):
                with (
                    self.subTest(method=method, code=code),
                    responses.RequestsMock(registry=OrderedRegistry) as rsps,
                ):
                    self.responses_add_test_endpoint(
                        method=method, status=code, rsps=rsps
                    )
                    self.responses_add_test_endpoint(
                        method=method, status=requests.codes.ok, rsps=rsps
                    )
                    debusine_client._api_request(
                        method, self.api_url, SampleResponse
                    )
                    rsps.reset()

    def test_api_request_dont_retry_error(self) -> None:  # pragma: no cover
        """Many HTTP error responses are not retried, test a few."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)

        for code in (
            requests.codes.bad_request,  # 400
            requests.codes.unauthorized,  # 401
            requests.codes.forbidden,  # 403
            # Enabled by default in Retry:
            requests.codes.request_entity_too_large,  # 413
        ):
            with (
                self.subTest(code=code),
                responses.RequestsMock(registry=OrderedRegistry) as rsps,
            ):
                self.responses_add_test_endpoint(status=code, rsps=rsps)
                self.responses_add_test_endpoint(
                    status=requests.codes.ok, rsps=rsps
                )
                with self.assertRaises(
                    (
                        exceptions.ClientForbiddenError,
                        exceptions.UnexpectedResponseError,
                    )
                ):
                    debusine_client._api_request(
                        responses.GET, self.api_url, SampleResponse
                    )
                rsps.reset()

    @responses.activate
    def test_api_request_returns_not_json_raise_exception(self) -> None:
        """Raise UnexpectedResponseError if body is not a valid JSON."""
        responses.add(
            responses.GET,
            self.api_url,
            body='Something that is not JSON as client expects',
        )
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            rf"^Server \({self.api_url}\) did not return valid object. Error: ",
        ):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        self.assert_token_key_included_in_all_requests(self.token)

    def test_api_request_create_and_use_session(self) -> None:
        """_api_request method create a session, uses and reuses it."""
        session_patch = mock.patch("requests.Session", autospec=True)
        session_mock = session_patch.start()
        session_mock.return_value.headers = {}
        self.addCleanup(session_patch.stop)

        token = "token"
        debusine_client = DebusineHttpClient(self.api_url, token)

        # Lazy creation of the session: not created yet
        self.assertIsNone(debusine_client._session)

        # Make a call (the HTTP GET is done, but it failed)
        with self.assertRaises(exceptions.UnexpectedResponseError):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        session_mock.assert_called_once()
        session_mock.return_value.get.assert_called_once_with(
            self.api_url, headers={"Token": token}, timeout=ANY
        )

        session = debusine_client._session
        # Make a call (the HTTP GET is done, but it failed)
        with self.assertRaises(exceptions.UnexpectedResponseError):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        # Same session reused
        self.assertIs(debusine_client._session, session)

    def test_api_request_no_token(self) -> None:
        """_api_request sends no Token header if the client has no token."""
        session_patch = mock.patch("requests.Session", autospec=True)
        session_mock = session_patch.start()
        session_mock.return_value.headers = {}
        self.addCleanup(session_patch.stop)

        debusine_client = DebusineHttpClient(self.api_url, None)

        # Make a call (the HTTP GET is done, but it failed)
        with self.assertRaises(exceptions.UnexpectedResponseError):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        session_mock.assert_called_once()
        session_mock.return_value.get.assert_called_once_with(
            self.api_url, headers={}, timeout=ANY
        )

    def test_api_request_scope(self) -> None:
        """_api_request sends X-Debusine-Scope header."""
        session_patch = mock.patch("requests.Session", autospec=True)
        session_mock = session_patch.start()
        session_mock.return_value.headers = {}
        self.addCleanup(session_patch.stop)

        debusine_client = DebusineHttpClient(self.api_url, None, scope="test")

        # Lazy creation of the session: not created yet
        self.assertIsNone(debusine_client._session)

        # Make a call (the HTTP GET is done, but it failed)
        with self.assertRaises(exceptions.UnexpectedResponseError):
            debusine_client._api_request("GET", self.api_url, SampleResponse)

        session_mock.assert_called_once()
        session_mock.return_value.get.assert_called_once_with(
            self.api_url, headers={"X-Debusine-Scope": "test"}, timeout=ANY
        )

    @responses.activate
    def test_api_request_submit_data_different_status_codes(self) -> None:
        """_api_request() return requests.codes.ok or created."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        data = {"foo": "bar"}

        for status_code in [requests.codes.ok, requests.codes.created]:
            responses.reset()
            responses.add(
                responses.POST,
                self.api_url,
                status=status_code,
                json={'id': 1},
            )

            if status_code == requests.codes.ok:
                expected_statuses = None
            else:
                expected_statuses = [status_code]

            debusine_client._api_request(
                "POST",
                self.api_url,
                SampleResponse,
                data=data,
                expected_statuses=expected_statuses,
            )

            self.assert_token_key_included_in_all_requests(self.token)
            assert responses.calls[0].request.body is not None
            self.assertEqual(json.loads(responses.calls[0].request.body), data)

    @responses.activate
    def test_api_request_ok_no_content_type(self) -> None:
        """_api_request() handles a 200 response with no Content-Type."""

        class EmptyResponse(responses.BaseResponse):
            """A response without even a Content-Type."""

            @override
            def get_response(
                self, request: requests.PreparedRequest
            ) -> HTTPResponse:
                headers = HTTPHeaderDict()
                body = BytesIO()
                # See comments in responses._form_response for the mypy
                # suppressions here.
                orig_response = HTTPResponse(
                    body=body,
                    msg=headers,  # type: ignore[arg-type]
                )
                return HTTPResponse(
                    status=requests.codes.ok,
                    body=body,
                    headers=headers,
                    original_response=orig_response,  # type: ignore[arg-type]
                    request_method=request.method,
                )

        debusine_client = DebusineHttpClient(self.api_url, self.token)
        responses.add(EmptyResponse("PUT", self.api_url))

        debusine_client._api_request("PUT", self.api_url, None)

    @responses.activate
    def test_debusine_error_details_combinations_everything(self) -> None:
        """Test combinations of status and body for _debusine_error_details."""
        json_with_error_message = {'title': 'error message'}
        no_json = 'There is no JSON'

        content_type_problem = 'application/problem+json'
        content_type_not_problem = 'application/json'

        for body, expected, content_type in (
            (
                json_with_error_message,
                json_with_error_message,
                content_type_problem,
            ),
            (json_with_error_message, None, content_type_not_problem),
            (no_json, None, content_type_problem),
            (json_with_error_message, None, content_type_not_problem),
            (no_json, None, content_type_problem),
        ):
            with self.subTest(body=body):
                responses.reset()
                kwargs: dict[str, Any] = {'content_type': content_type}

                if isinstance(body, dict):
                    kwargs['json'] = body
                else:
                    kwargs['body'] = body

                responses.add(responses.GET, self.api_url, **kwargs)
                response = requests.get(self.api_url)

                error = DebusineHttpClient._debusine_problem(response)

                self.assertEqual(error, expected)

    @responses.activate
    def test_api_request_raise_client_connection_error(self) -> None:
        """Raise ClientConnectionError for RequestException/ConnectionError."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)

        for exception in (
            requests.exceptions.RequestException,
            ConnectionError,
        ):
            with self.subTest(exception=repr(exception)):
                responses.add(
                    responses.GET,
                    self.api_url,
                    body=exception(),
                )

                with self.assertRaisesRegex(
                    exceptions.ClientConnectionError,
                    f"^Cannot connect to {self.api_url}. Error: $",
                ):
                    debusine_client._api_request(
                        "GET", self.api_url, SampleResponse
                    )

    @responses.activate
    def test_iter_paginated_get_iterates(self) -> None:
        """iter_paginated_get() iterates across pages."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        etag = '"some-etag"'
        self.responses_add_paginated_responses(
            self.api_url,
            length=4,
            page_size=3,
            etag=etag,
        )

        iter_ = debusine_client.iter_paginated_get("", SampleResponse)
        for i in range(4):
            obj = next(iter_)
            self.assertEqual(obj.id, i)
        self.assertEqual(obj.page, 2)

        # Received the etag
        response = responses.calls[0].response
        assert isinstance(response, requests.models.Response)
        assert response.headers is not None
        self.assertEqual(response.headers["ETag"], etag)

    @responses.activate
    def test_iter_paginated_get_terminates(self) -> None:
        """iter_paginated_get() terminates at the end."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        self.responses_add_paginated_responses(
            self.api_url, length=4, page_size=3
        )

        response = list(debusine_client.iter_paginated_get("", SampleResponse))
        self.assertEqual(
            response,
            [
                SampleResponse(id=0, page=1),
                SampleResponse(id=1, page=1),
                SampleResponse(id=2, page=1),
                SampleResponse(id=3, page=2),
            ],
        )

    @responses.activate
    def test_get_single_page_with_headers(self) -> None:
        debusine_client = DebusineHttpClient(self.api_url, self.token)

        etag = "etag-1"

        expected = self.responses_add_paginated_responses(
            self.api_url,
            length=5,
            page_size=5,
            etag=etag,
        )

        content, headers = debusine_client.get_single_page_with_headers(
            "", SampleResponse
        )

        self.assertEqual(content, expected)
        self.assertEqual(
            headers, {"Content-Type": "application/json", "ETag": etag}
        )

    @responses.activate
    def test_get_single_page_with_headers_has_next(self) -> None:
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        self.responses_add_paginated_responses(
            self.api_url, length=10, page_size=2
        )

        with self.assertRaisesRegex(
            UnexpectedResponseError,
            re.escape(
                "Expected a single paginated response page from "
                f"{self.api_url}, but server returned additional pages"
            ),
        ):
            debusine_client.get_single_page_with_headers("", SampleResponse)

    @responses.activate
    def test_get_single_page_with_headers_invalid_object(self) -> None:
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        responses.add(
            responses.GET,
            self.api_url,
            json={
                "count": 1,
                "results": [{"a": "b"}],
            },
        )

        with self.assertRaisesRegex(
            UnexpectedResponseError,
            re.escape(f"Server ({self.api_url}) did not return valid object"),
        ):
            debusine_client.get_single_page_with_headers("", SampleResponse)

    @responses.activate
    def test_iter_paginated_get_handles_empty_set(self) -> None:
        """iter_paginated_get() handles an empty resultset."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        self.responses_add_paginated_responses(
            self.api_url, length=0, page_size=3
        )

        response = list(debusine_client.iter_paginated_get("", SampleResponse))
        self.assertEqual(response, [])

    @responses.activate
    def test_iter_paginated_get_invalid_pagination_structure(self) -> None:
        """iter_paginated_get() handles an invalid pagination response."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        responses.add(
            responses.GET,
            self.api_url,
            json={"foo": "bar"},
        )
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            rf"^Server \({self.api_url}\) did not return valid object. Error: ",
        ):
            next(debusine_client.iter_paginated_get("", SampleResponse))

    @responses.activate
    def test_iter_paginated_get_invalid_response(self) -> None:
        """iter_paginated_get() handles an invalid object in responses."""
        debusine_client = DebusineHttpClient(self.api_url, self.token)
        responses.add(
            responses.GET,
            self.api_url,
            json={
                "count": 12,
                "results": [{"foo": "bar"}],
            },
        )
        with self.assertRaisesRegex(
            exceptions.UnexpectedResponseError,
            rf"^Server \({self.api_url}\) did not return valid object. Error: ",
        ):
            next(debusine_client.iter_paginated_get("", SampleResponse))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to manage scopes.

Note: to make commands easier to be invoked from Ansible, we take care to make
them idempotent.
"""

from argparse import BooleanOptionalAction
from collections.abc import Callable
from typing import Any, NoReturn, cast, override

import rich
from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import transaction
from django.db.models import F
from rich.table import Table

from debusine.db.context import context
from debusine.db.models import (
    DEFAULT_FILE_STORE_NAME,
    File,
    FileStore,
    FileStoreInScope,
    Group,
    Scope,
)
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import get_scope


class Command(DebusineBaseCommand):
    """Command to manage scopes."""

    help = "Manage scopes"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        create = subparsers.add_parser("create", help="Ensure a scope exists")
        create.add_argument("name", help="Name for the new scope")
        create.add_argument(
            "--with-owners-group",
            type=str,
            nargs="?",
            metavar="name",
            const="Owners",
            help="Ensure the scope has the named owners group"
            " (name defaults to 'Owners')",
        )
        create.add_argument(
            "--label",
            type=str,
            help="Scope label. Defaults to capitalized name",
        )
        create.add_argument(
            "--icon",
            type=str,
            help="Optional scope icon, as a path resolved"
            " by the static tag in templates",
        )
        create.add_argument(
            "--file-store",
            dest="file_store_name",
            metavar="NAME",
            default=DEFAULT_FILE_STORE_NAME,
            help=(
                "Name of the initial file store to use for this scope"
                f" (default: {DEFAULT_FILE_STORE_NAME})"
            ),
        )

        rename = subparsers.add_parser("rename", help="Rename a scope")
        rename.add_argument("scope", help="Scope to rename")
        rename.add_argument("name", help="New name for the scope")

        add_file_store = subparsers.add_parser(
            "add_file_store", help="Add a file store to a scope"
        )
        add_file_store.add_argument("scope_name", help="Scope to manage")
        add_file_store.add_argument("file_store_name", help="File store to add")

        edit_file_store = subparsers.add_parser(
            "edit_file_store", help="Edit a file store in a scope"
        )
        edit_file_store.add_argument("scope_name", help="Scope to manage")
        edit_file_store.add_argument(
            "file_store_name", help="File store to edit"
        )

        remove_file_store = subparsers.add_parser(
            "remove_file_store", help="Remove a file store from a scope"
        )
        remove_file_store.add_argument("scope_name", help="Scope to manage")
        remove_file_store.add_argument(
            "file_store_name", help="File store to remove"
        )

        for subparser in (add_file_store, edit_file_store):
            subparser.add_argument(
                "--upload-priority",
                help=(
                    "Upload priority for the selected file store (use the "
                    "empty string to unset)"
                ),
            )
            subparser.add_argument(
                "--download-priority",
                help=(
                    "Download priority for the selected file store (use the "
                    "empty string to unset)"
                ),
            )
            subparser.add_argument(
                "--populate",
                action=BooleanOptionalAction,
                help=(
                    "Ensure that this store has a copy of all files in the "
                    "scope"
                ),
            )
            subparser.add_argument(
                "--drain",
                action=BooleanOptionalAction,
                help=(
                    "Schedule all files in this scope to be moved to some "
                    "other store in the same scope"
                ),
            )
            subparser.add_argument(
                "--drain-to",
                help=(
                    "Constrain --drain to use the store with this name in the "
                    "scope"
                ),
            )
            subparser.add_argument(
                "--read-only",
                action=BooleanOptionalAction,
                help="Do not add new files to this store",
            )
            subparser.add_argument(
                "--write-only",
                action=BooleanOptionalAction,
                help="Do not read files from this store",
            )
            subparser.add_argument(
                "--soft-max-size",
                help=(
                    "The number of bytes that the file store can hold for this "
                    "scope (use the empty string to unset)"
                ),
            )

        show = subparsers.add_parser("show", help="Show a scope")
        show.add_argument("scope", help="Scope to show")

    def get_file_store(self, name: str) -> FileStore:
        """Look up a file store by name."""
        try:
            return FileStore.objects.get(name=name)
        except FileStore.DoesNotExist:
            raise CommandError(f"File store {name!r} not found", returncode=3)

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        if func is None:
            raise CommandError(
                f"Action {options['action']!r} not found", returncode=3
            )

        func(*args, **options)

    @context.disable_permission_checks()
    def handle_create(
        self,
        *,
        name: str,
        with_owners_group: str | None = None,
        label: str | None,
        icon: str | None,
        file_store_name: str,
        **options: Any,
    ) -> NoReturn:
        """
        Create a scope, initialized with an Owners group.

        This is idempotent, and it makes sure the named scope exists, it has an
        "Owners" group, and that group has the ADMIN role on the scope.  It
        also makes sure that the named scope uses the given file store,
        defaulting to upload and download priorities of 100.
        """
        label = label or name.capitalize()
        icon = icon or ""
        with transaction.atomic():
            scope, _ = Scope.objects.get_or_create(
                name=name, defaults={"label": label, "icon": icon}
            )
            if with_owners_group:
                # Make sure the scope has an Owners group
                admin_group, _ = Group.objects.get_or_create(
                    scope=scope, name=with_owners_group
                )
                admin_group.assign_role(scope, "owner")

            if scope.label != label:
                scope.label = label
            if scope.icon != icon:
                scope.icon = icon

            file_store = self.get_file_store(file_store_name)
            scope.file_stores.add(
                file_store,
                through_defaults={
                    "upload_priority": 100,
                    "download_priority": 100,
                },
            )

            try:
                scope.full_clean()
            except ValidationError as exc:
                self.stderr.write("New scope would be invalid:")
                for field, errors in exc.message_dict.items():
                    for error in errors:
                        self.stderr.write(f"* {field}: {error}")
                raise SystemExit(3)
            scope.save()

        raise SystemExit(0)

    def handle_rename(
        self, *, scope: str, name: str, **options: Any
    ) -> NoReturn:
        """Rename a scope."""
        s = get_scope(scope)

        if name == s.name:
            raise SystemExit(0)

        s.name = name
        try:
            s.full_clean()
        except ValidationError as exc:
            self.stderr.write("Renamed scope would be invalid:")
            for field, errors in exc.message_dict.items():
                for error in errors:
                    self.stderr.write(f"* {field}: {error}")
            raise SystemExit(3)

        s.save()

        raise SystemExit(0)

    def handle_add_file_store(
        self,
        *,
        scope_name: str,
        file_store_name: str,
        upload_priority: str | None = None,
        download_priority: str | None = None,
        populate: bool | None = None,
        drain: bool | None = None,
        drain_to: str | None = None,
        read_only: bool | None = None,
        write_only: bool | None = None,
        soft_max_size: str | None = None,
        **options: Any,
    ) -> NoReturn:
        """Add a file store to a scope."""
        scope = get_scope(scope_name)
        file_store = self.get_file_store(file_store_name)
        through_defaults = {
            "upload_priority": upload_priority or None,
            "download_priority": download_priority or None,
            "populate": populate or False,
            "drain": drain or False,
            "drain_to": drain_to or None,
            "read_only": read_only or False,
            "write_only": write_only or False,
            "soft_max_size": soft_max_size or None,
        }
        scope.file_stores.add(file_store, through_defaults=through_defaults)

        raise SystemExit(0)

    def handle_edit_file_store(
        self,
        *,
        scope_name: str,
        file_store_name: str,
        upload_priority: str | None = None,
        download_priority: str | None = None,
        populate: bool | None = None,
        drain: bool | None = None,
        drain_to: str | None = None,
        read_only: bool | None = None,
        write_only: bool | None = None,
        soft_max_size: str | None = None,
        **options: Any,
    ) -> NoReturn:
        """Edit properties of a file store in a scope."""
        scope = get_scope(scope_name)
        try:
            file_store_extra = scope.filestoreinscope_set.get(
                file_store__name=file_store_name
            )
        except FileStoreInScope.DoesNotExist:
            raise CommandError(
                f"Scope {scope.name!r} has no file store named "
                f"{file_store_name!r}",
                returncode=3,
            )
        if upload_priority is not None:
            file_store_extra.upload_priority = upload_priority or None
        if download_priority is not None:
            file_store_extra.download_priority = download_priority or None
        if populate is not None:
            file_store_extra.populate = populate
        if drain is not None:
            file_store_extra.drain = drain
        if drain_to is not None:
            file_store_extra.drain_to = drain_to or None
        if read_only is not None:
            file_store_extra.read_only = read_only
        if write_only is not None:
            file_store_extra.write_only = write_only
        if soft_max_size is not None:
            file_store_extra.soft_max_size = soft_max_size or None
        file_store_extra.save()

        raise SystemExit(0)

    def handle_remove_file_store(
        self, scope_name: str, file_store_name: str, **options: Any
    ) -> NoReturn:
        """Remove a file store from a scope."""
        scope = get_scope(scope_name)
        try:
            file_store = scope.file_stores.get(name=file_store_name)
        except FileStore.DoesNotExist:
            raise CommandError(
                f"Scope {scope.name!r} has no file store named "
                f"{file_store_name!r}",
                returncode=3,
            )
        files_in_this_store = File.objects.filter(
            artifact__workspace__scope=scope, filestore=file_store
        )
        files_in_other_stores = File.objects.filter(
            artifact__workspace__scope=scope,
            filestore__in=scope.file_stores.exclude(pk=file_store.pk),
        )
        if files_in_this_store.exclude(pk__in=files_in_other_stores).exists():
            raise CommandError(
                f"Scope {scope.name!r} contains some files that are only in "
                f"file store {file_store_name!r}; drain it first",
                returncode=3,
            )
        scope.file_stores.remove(file_store)

        raise SystemExit(0)

    def handle_show(self, scope: str, **options: Any) -> NoReturn:
        """Show a scope."""
        s = get_scope(scope)

        # TODO: This layout is a bit odd: we're trying to fit everything
        # into a single table despite the true structure being nested.  It
        # does the job for now and is compatible with TabularOutputTests.
        # TODO: Add an option for YAML output.
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("Name")
        table.add_column("Label")
        table.add_column("File store name")
        table.add_column("Backend")
        table.add_column("Upload priority")
        table.add_column("Download priority")
        table.add_column("# Files")
        for i, file_store_extra in enumerate(
            s.filestoreinscope_set.order_by(
                F("upload_priority").desc(nulls_last=True),
                F("download_priority").desc(nulls_last=True),
            )
        ):
            file_store = file_store_extra.file_store
            files_in_this_store = File.objects.filter(
                artifact__workspace__scope=s, filestore=file_store
            ).distinct()
            table.add_row(
                s.name if i == 0 else None,
                s.label if i == 0 else None,
                file_store.name,
                file_store.backend,
                (
                    None
                    if file_store_extra.upload_priority is None
                    else str(file_store_extra.upload_priority)
                ),
                (
                    None
                    if file_store_extra.download_priority is None
                    else str(file_store_extra.download_priority)
                ),
                str(files_in_this_store.count()),
            )
        rich.print(table)

        raise SystemExit(0)
